"""Module providing objects for working with reference reflections and orientations."""
import dataclasses
from typing import List, Tuple, Union

from diffcalc.hkl.geometry import Position


@dataclasses.dataclass
class Reflection:
    """Class containing reference reflection information.

    Attributes
    ----------
    h: float
        h miller index.
    k: float
        k miller index.
    l: float
        l miller index.
    pos: Position
        Diffractometer position object.
    energy: float
        Beam energy in keV.
    tag: str
        Identifying tag for the reflection.
    """

    h: float
    k: float
    l: float
    pos: Position
    energy: float
    tag: str

    def __post_init__(self):
        """Check input argument types.

        Raises
        ------
        TypeError
            If pos argument has invalid type.
        """
        if not isinstance(self.pos, Position):
            raise TypeError(f"Invalid position object type {type(self.pos)}.")

    @property
    def astuple(
        self,
    ) -> Tuple[
        Tuple[float, float, float],
        Tuple[float, float, float, float, float, float],
        float,
        str,
    ]:
        """Return reference reflection data as tuple.

        Returns
        -------
        Tuple[Tuple[float, float, float],
              Tuple[float, float, float, float, float, float],
              float,
              str]
            Tuple containing miller indices, position object, energy and
            reflection tag.
        """
        h, k, l, pos, en, tag = dataclasses.astuple(self)
        return (h, k, l), pos.astuple, en, tag


class ReflectionList:
    """Class containing collection of reference reflections.

    Attributes
    ----------
    reflections: List[Reflection]
        List containing reference reflections.
    """

    def __init__(self, reflections=None):
        self.reflections: List[Reflection] = reflections if reflections else []

    def get_tag_index(self, tag: str) -> int:
        """Get a reference reflection index.

        Get a reference reflection index for the provided reflection tag.

        Parameters
        ----------
        tag : str
            identifying tag for the reflection

        Returns
        -------
        int:
            The reference reflection index.

        Raises
        ------
        ValueError
            If tag not found in reflection list.
        """
        _tag_list = [ref.tag for ref in self.reflections]
        num = _tag_list.index(tag)
        return num

    def add_reflection(
        self, hkl: Tuple[float, float, float], pos: Position, energy: float, tag: str
    ) -> None:
        """Add a reference reflection.

        Adds a reference reflection object to the reflection list.

        Parameters
        ----------
        hkl : Tuple[float, float, float]
            Miller indices of the reflection
        pos: Position
            Object representing diffractometer angles
        energy : float
            Energy of the x-ray beam.
        tag : str
            Identifying tag for the reflection.
        """
        self.reflections += [Reflection(*hkl, pos, energy, tag)]

    def edit_reflection(
        self,
        idx: Union[str, int],
        hkl: Tuple[float, float, float],
        pos: Position,
        energy: float,
        tag: str,
    ) -> None:
        """Change a reference reflection.

        Changes the reference reflection object in the reflection list.

        Parameters
        ----------
        idx : Union[str, int]
            Index or tag of the reflection to be changed
        hkl : Tuple[float,float,float]
            Miller indices of the reflection
        position: Position
            Object representing diffractometer angles.
        energy : float
            Energy of the x-ray beam.
        tag : str
            Identifying tag for the reflection.

        Raises
        ------
        ValueError
            Reflection with specified tag not found.
        IndexError
            Reflection with specified index not found.
        """
        if isinstance(idx, str):
            num = self.get_tag_index(idx)
        else:
            num = idx - 1
        if isinstance(pos, Position):
            self.reflections[num] = Reflection(*hkl, pos, energy, tag)
        else:
            raise TypeError("Invalid position parameter type")

    def get_reflection(self, idx: Union[str, int]) -> Reflection:
        """Get a reference reflection.

        Get aon object representing reference reflection.

        Parameters
        ----------
        idx : Union[str, int]
            Index or tag of the reflection.

        Returns
        -------
        Reflection
            Object representing reference reflection

        Raises
        ------
        ValueError
            Reflection with the requested index/tan not present.
        IndexError
            Reflection with specified index not found.
        """
        if isinstance(idx, str):
            num = self.get_tag_index(idx)
        else:
            num = idx - 1
        return self.reflections[num]

    def remove_reflection(self, idx: Union[str, int]) -> None:
        """Delete a reference reflection.

        Parameters
        ----------
        idx : Union[str, int]
            Index or tag of the deleted reflection.

        Raises
        ------
        ValueError
            Reflection with the requested index/tag not present.
        IndexError
            Reflection with specified index not found.
        """
        if isinstance(idx, str):
            num = self.get_tag_index(idx)
        else:
            num = idx - 1
        del self.reflections[num]

    def swap_reflections(self, idx1: Union[str, int], idx2: Union[str, int]) -> None:
        """Swap indices of two reference reflections.

        Parameters
        ----------
        idx1 : Union[str, int]
            Index or tag of the first reflection to be swapped.
        idx2 : Union[str, int]
            Index or tag of the second reflection to be swapped.

        Raises
        ------
        ValueError
            Reflection with the requested index/tag not present.
        IndexError
            Reflection with specified index not found.
        """
        if isinstance(idx1, str):
            num1 = self.get_tag_index(idx1)
        else:
            num1 = idx1 - 1
        if isinstance(idx2, str):
            num2 = self.get_tag_index(idx2)
        else:
            num2 = idx2 - 1
        orig1 = self.reflections[num1]
        self.reflections[num1] = self.reflections[num2]
        self.reflections[num2] = orig1

    def __len__(self) -> int:
        """Return number of reference reflections in the list.

        Returns
        -------
        int
            Number of reference reflections.
        """
        return len(self.reflections)

    def __str__(self) -> str:
        """Represent the reference reflection list as a string.

        Returns
        -------
        str
            Table containing list of all reflections.
        """
        return "\n".join(self._str_lines())

    def _str_lines(self) -> List[str]:
        """Table with reference reflection data.

        Returns
        -------
        List[str]
            List containing reference reflection table rows.
        """
        axes = tuple(fd.upper() for fd in Position.fields)
        if not self.reflections:
            return ["   <<< none specified >>>"]

        lines = []

        fmt = "     %6s %5s %5s %5s  " + "%8s " * len(axes) + " TAG"
        header_values = ("ENERGY", "H", "K", "L") + axes
        lines.append(fmt % header_values)

        for n in range(1, len(self.reflections) + 1):
            ref_tuple = self.get_reflection(n)
            (h, k, l), pos, energy, tag = ref_tuple.astuple
            if tag is None:
                tag = ""
            fmt = "  %2d %6.3f % 4.2f % 4.2f % 4.2f  " + "% 8.4f " * len(axes) + " %s"
            values = (n, energy, h, k, l) + pos + (tag,)
            lines.append(fmt % values)
        return lines


@dataclasses.dataclass
class Orientation:
    """Class containing reference orientation information.

    Attributes
    ----------
    h: float
        h miller index.
    k: float
        k miller index.
    l: float
        l miller index.
    x: float
        x coordinate in laboratory system.
    y: float
        y coordinate in laboratory system.
    z: float
        z coordinate in laboratory system.
    pos: Position
        Diffractometer position object.
    tag: str
        Identifying tag for the orientation.
    """

    h: float
    k: float
    l: float
    x: float
    y: float
    z: float
    pos: Position
    tag: str

    def __post_init__(self):
        """Check input argument types.

        Raises
        ------
        TypeError
            If pos argument has invalid type.
        """
        if not isinstance(self.pos, Position):
            raise TypeError(f"Invalid position object type {type(self.pos)}.")

    @property
    def astuple(
        self,
    ) -> Tuple[
        Tuple[float, float, float],
        Tuple[float, float, float],
        Tuple[float, float, float, float, float, float],
        str,
    ]:
        """Return reference orientation data as tuple.

        Returns
        -------
        Tuple[Tuple[float, float, float],
              Tuple[float, float, float],
              Tuple[float, float, float, float, float, float],
              str]
            Tuple containing miller indices, laboratory frame coordinates,
            position object and orientation tag.
        """
        h, k, l, x, y, z, pos, tag = dataclasses.astuple(self)
        return (h, k, l), (x, y, z), pos.astuple, tag


class OrientationList:
    """Class containing collection of reference orientations.

    Attributes
    ----------
    reflections: List[Orientation]
        List containing reference orientations.
    """

    def __init__(self, orientations=None):
        self.orientations: List[Orientation] = orientations if orientations else []

    def get_tag_index(self, tag: str) -> int:
        """Get a reference orientation index.

        Get a reference orientation index for the provided orientation tag.

        Parameters
        ----------
        tag : str
            identifying tag for the orientation

        Returns
        -------
        int:
            The reference orientation index.

        Raises
        ------
        ValueError
            If tag not found in orientations list.
        """
        _tag_list = [orient.tag for orient in self.orientations]
        num = _tag_list.index(tag)
        return num

    def add_orientation(
        self,
        hkl: Tuple[float, float, float],
        xyz: Tuple[float, float, float],
        pos: Position,
        tag: str,
    ) -> None:
        """Add a reference orientation.

        Adds a reference orientation in the external diffractometer
        coordinate system.

        Parameters
        ----------
        hkl : Tuple[float, float, float]
            Miller index of the reference orientation.
        xyz : Tuple[float, float, float]
            Laboratory frame coordinates of the reference orientation.
        position: Position
            Object representing diffractometer position.
        tag : str
            identifying tag for the orientation.
        """
        if isinstance(pos, Position):
            self.orientations += [Orientation(*hkl, *xyz, pos, tag)]
        else:
            raise TypeError("Invalid position parameter type")

    def edit_orientation(
        self,
        idx: Union[str, int],
        hkl: Tuple[float, float, float],
        xyz: Tuple[float, float, float],
        pos: Position,
        tag: str,
    ) -> None:
        """Change a reference orientation.

        Changes a reference orientation in the external diffractometer
        coordinate system.

        Parameters
        ----------
        idx : str or int
            Index or tag of the orientation to be changed.
        hkl : Tuple[float, float, float]
            Miller indices of the reference orientation.
        xyz : :Tuple[float, float, float]
            Laboratory frame coordinates of the reference orientation.
        pos: Position
            Object representing diffractometer position.
        tag : str
            Identifying tag for the orientation.

        Raises
        ------
        ValueError
            Orientation with specified tag not found.
        IndexError
            Orientation with specified index not found.
        """
        if isinstance(idx, str):
            num = self.get_tag_index(idx)
        else:
            num = idx - 1
        if isinstance(pos, Position):
            self.orientations[num] = Orientation(*hkl, *xyz, pos, tag)
        else:
            raise TypeError(f"Invalid position parameter type {type(pos)}")

    def get_orientation(self, idx: Union[str, int]) -> Orientation:
        """Get a reference orientation.

        Get an object representing reference orientation.

        Parameters
        ----------
        idx : Union[str, int]
            Index or tag of the orientation.

        Returns
        -------
        Orientation
            Object representing reference orientation.

        Raises
        ------
        ValueError
            Orientation with the requested index/tag not present.
        IndexError
            Orientation with specified index not found.
        """
        if isinstance(idx, str):
            num = self.get_tag_index(idx)
        else:
            num = idx - 1
        return self.orientations[num]

    def remove_orientation(self, idx: Union[str, int]) -> None:
        """Delete a reference orientation.

        Parameters
        ----------
        idx : Union[str, int]
            Index or tag of the deteled orientation.

        Raises
        ------
        ValueError
            Orientation with the requested index/tag not present.
        IndexError
            Orientation with specified index not found.
        """
        if isinstance(idx, str):
            num = self.get_tag_index(idx)
        else:
            num = idx - 1
        del self.orientations[num]

    def swap_orientations(self, idx1: Union[str, int], idx2: Union[str, int]) -> None:
        """Swap indices of two reference orientations.

        Parameters
        ----------
        idx1 : Union[str, int]
            Index or tag of the first orientation to be swapped.
        idx2 : Union[str, int]
            Index or tag of the second orientation to be swapped.

        Raises
        ------
        ValueError
            Orientation with the requested index/tag not present.
        IndexError
            Orientation with specified index not found.
        """
        if isinstance(idx1, str):
            num1 = self.get_tag_index(idx1)
        else:
            num1 = idx1 - 1
        if isinstance(idx2, str):
            num2 = self.get_tag_index(idx2)
        else:
            num2 = idx2 - 1
        orig1 = self.orientations[num1]
        self.orientations[num1] = self.orientations[num2]
        self.orientations[num2] = orig1

    def __len__(self) -> int:
        """Return number of reference orientations in the list.

        Returns
        -------
        int
            Number of reference orientationss.
        """
        return len(self.orientations)

    def __str__(self) -> str:
        """Represent the reference orientations list as a string.

        Returns
        -------
        str
            Table containing list of all orientations.
        """
        return "\n".join(self._str_lines())

    def _str_lines(self) -> List[str]:
        """Table with reference orientations data.

        Returns
        -------
        List[str]
            List containing reference orientations table rows.
        """
        axes = tuple(fd.upper() for fd in Position.fields)
        if not self.orientations:
            return ["   <<< none specified >>>"]

        lines = []

        str_format = "     %5s %5s %5s   %5s %5s %5s  " + "%8s " * len(axes) + " TAG"
        header_values = ("H", "K", "L", "X", "Y", "Z") + axes
        lines.append(str_format % header_values)

        for n in range(1, len(self.orientations) + 1):
            orient = self.get_orientation(n)
            (h, k, l), (x, y, z), angles, tag = orient.astuple
            if tag is None:
                tag = ""
            str_format = (
                "  %2d % 4.2f % 4.2f % 4.2f  "
                + "% 4.2f % 4.2f % 4.2f  "
                + "% 8.4f " * len(axes)
                + " %s"
            )
            values = (n, h, k, l, x, y, z) + angles + (tag,)
            lines.append(str_format % values)
        return lines



###
# Copyright 2008-2011 Diamond Light Source Ltd.
# This file is part of Diffcalc.
#
# Diffcalc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Diffcalc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Diffcalc.  If not, see <http://www.gnu.org/licenses/>.
###

import pytest
from diffcalc.ub.calc import ReferenceVector
from diffcalc.util import DiffcalcException
from numpy import array

from tests.tools import assert_2darray_almost_equal


@pytest.fixture
def reference():
    return ReferenceVector((0, 0, 1), False)


@pytest.mark.parametrize(
    ("vector"),
    [
        array([[1], [2], [3]]),
        pytest.param([[1], [2]], marks=pytest.mark.xfail(raises=DiffcalcException)),
        pytest.param(
            array([[1], [2]]), marks=pytest.mark.xfail(raises=DiffcalcException)
        ),
    ],
)
def test_from_as_array(reference, vector):
    reference.set_array(vector)
    assert reference.n_ref == tuple(vector.T[0])
    result = reference.get_array()
    assert_2darray_almost_equal(vector, result)


def test_default_n_phi(reference):
    assert_2darray_almost_equal(reference.get_array(), array([[0], [0], [1]]).tolist())


def test__str__with_phi_configured(reference):
    print(reference)


def test__str__with_hkl_configured(reference):
    reference = ReferenceVector((0, 1, 1), True)
    print(reference)


def test_n_phi_from_hkl_with_unity_matrix_001(reference):
    reference = ReferenceVector((0, 0, 1), True)
    assert_2darray_almost_equal(reference.get_array(), array([[0], [0], [1]]))


def test_n_phi_from_hkl_with_unity_matrix_010(reference):
    reference = ReferenceVector((0, 1, 0), True)
    assert_2darray_almost_equal(reference.get_array(), array([[0], [1], [0]]))



###
# Copyright 2008-2011 Diamond Light Source Ltd.
# This file is part of Diffcalc.
#
# Diffcalc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Diffcalc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Diffcalc.  If not, see <http://www.gnu.org/licenses/>.
###

from diffcalc.hkl.geometry import Position

from tests.tools import degrees_equivalent


class TestUtils:
    def testDegreesEqual(self):
        tol = 0.001
        assert degrees_equivalent(1, 1, tol)
        assert degrees_equivalent(1, -359, tol)
        assert degrees_equivalent(359, -1, tol)
        assert not degrees_equivalent(1.1, 1, tol)
        assert not degrees_equivalent(1.1, -359, tol)
        assert not degrees_equivalent(359.1, -1, tol)


class TestPosition:
    def testCompare(self):
        # Test the compare method
        pos1 = Position(1, 2, 3, 4, 5, 6)
        pos2 = Position(1.1, 2.1, 3.1, 4.1, 5.1, 6.1)

        assert pos1 == pos1
        assert pos1 != pos2



"""Routines for calculating miller indices and diffractometer positions.

Module implementing calculations based on UB matrix data and diffractometer
constraints.
"""
from copy import copy
from itertools import product
from math import acos, asin, atan, atan2, cos, degrees, isnan, pi, sin, sqrt, tan
from typing import Dict, Iterator, List, Optional, Tuple

import numpy as np
from diffcalc.hkl.geometry import (
    Position,
    get_rotation_matrices,
    rot_CHI,
    rot_ETA,
    rot_MU,
    rot_PHI,
)
from diffcalc.log import logging
from diffcalc.util import (
    SMALL,
    DiffcalcException,
    I,
    angle_between_vectors,
    bound,
    cross3,
    is_small,
    normalised,
    radians_equivalent,
    sign,
    x_rotation,
    y_rotation,
    z_rotation,
)
from numpy.linalg import inv, norm

logger = logging.getLogger("diffcalc.hkl.calc")


class HklCalculation:
    """Class for converting between miller indices and diffractometer position.

    Attributes
    ----------
    ubcalc: UBcalculation
        Reference to UBcalculation object containing UB matrix data.
    constraints:
        Reference to Constraints object containing diffractometer constraint settings.

    Methods
    -------
    get_hkl(pos: Position, wavelength: float) -> Tuple[float, float, float]
        Calculate miller indices corresponding to a diffractometer positions.
    get_virtual_angles(pos: Position, asdegrees: bool = True) -> Dict[str,float]
        Calculate pseudo-angles corresponding to a diffractometer position.
    """

    def __init__(self, ubcalc, constraints):
        self.ubcalc = ubcalc  # to get the UBMatrix
        self.constraints = constraints

    def __str__(self):
        """Return string representing class instance.

        Returns
        -------
        str:
            String representation of constraints table.
        """
        return self.constraints.__str__()

    def __repr_mode(self):
        return repr(self.constraints.asdict)

    def get_hkl(self, pos: Position, wavelength: float) -> Tuple[float, float, float]:
        """Calculate miller indices corresponding to a diffractometer positions.

        Parameters
        ----------
        pos: Position
            Diffractometer position

        Returns
        -------
        Tuple[float, float, float]
            Miller indices corresponding to the specified diffractometer
            position at the given wavelength.
        """
        pos_in_rad = Position.asradians(pos)
        [MU, DELTA, NU, ETA, CHI, PHI] = get_rotation_matrices(pos_in_rad)

        q_lab = (NU @ DELTA - I) @ np.array([[0], [2 * pi / wavelength], [0]])  # 12

        hkl = inv(self.ubcalc.UB) @ inv(PHI) @ inv(CHI) @ inv(ETA) @ inv(MU) @ q_lab

        return hkl[0, 0], hkl[1, 0], hkl[2, 0]

    def get_virtual_angles(
        self, pos: Position, asdegrees: bool = True
    ) -> Dict[str, float]:
        """Calculate pseudo-angles corresponding to a diffractometer position.

        Parameters
        ----------
        pos: Position
            Diffractometer position
        asdegrees: bool = True
            If True, return angles in degrees

        Returns
        -------
        Dict[str, float]
            Returns alpha, beta, betain, betaout, naz, psi, qaz, tau, theta and
            ttheta angles.
        """
        pos_in_rad = Position.asradians(pos)
        theta, qaz = self._theta_and_qaz_from_detector_angles(
            pos_in_rad.delta, pos_in_rad.nu
        )  # (19)

        [MU, DELTA, NU, ETA, CHI, PHI] = get_rotation_matrices(pos_in_rad)
        Z = MU @ ETA @ CHI @ PHI
        D = NU @ DELTA

        # Compute incidence and outgoing angles bin and betaout
        surf_nphi = Z @ self.ubcalc.surf_nphi
        kin = np.array([[0], [1], [0]])
        kout = D @ np.array([[0], [1], [0]])
        betain = angle_between_vectors(kin, surf_nphi) - pi / 2.0
        betaout = pi / 2.0 - angle_between_vectors(kout, surf_nphi)

        n_lab = Z @ self.ubcalc.n_phi
        alpha = asin(bound(-n_lab[1, 0]))
        naz = atan2(n_lab[0, 0], n_lab[2, 0])  # (20)

        cos_tau = cos(alpha) * cos(theta) * cos(naz - qaz) + sin(alpha) * sin(theta)
        tau = acos(bound(cos_tau))  # (23)

        # Compute Tau using the dot product directly (THIS ALSO WORKS)
        # q_lab = ( (NU @ DELTA - I ) @ np.array([[0],[1],[0]])
        # norm = norm(q_lab)
        # q_lab = np.array([[1],[0],[0]]) if norm == 0 else q_lab * (1/norm)
        # tau_from_dot_product = acos(bound(dot3(q_lab, n_lab)))

        sin_beta = 2 * sin(theta) * cos(tau) - sin(alpha)
        beta = asin(bound(sin_beta))  # (24)

        psi = next(self._calc_psi(alpha, theta, tau, qaz, naz))

        result = {
            "theta": theta,
            "ttheta": 2 * theta,
            "qaz": qaz,
            "alpha": alpha,
            "naz": naz,
            "tau": tau,
            "psi": psi,
            "beta": beta,
            "betain": betain,
            "betaout": betaout,
        }
        if asdegrees:
            result = {key: degrees(val) for key, val in result.items()}
        return result

    def get_position(
        self, h: float, k: float, l: float, wavelength: float, asdegrees: bool = True
    ) -> List[Tuple[Position, Dict[str, float]]]:
        """Calculate diffractometer position from miller indices and wavelength.

        The calculated positions and angles are verified by checking that they
        map to the requested miller indices.

        Parameters
        ----------
            h: float
                h miller index
            k: float
                k miller index
            l: float
                l miller index
            wavelength: float
                wavelength in Angstroms
            asdegrees: bool
                If True, return angles in degrees

        Returns
        -------
            List[Tuple[Position, Dict[str, float]]]
                List of all solutions matching the input miller indices that
                consists of pairs of diffractometer position object and virtual
                angles dictionary.
        """
        pos_virtual_angles_pairs = self._calc_hkl_to_position(h, k, l, wavelength)
        assert pos_virtual_angles_pairs
        results = []

        for pos, virtual_angles in pos_virtual_angles_pairs:
            self._verify_pos_map_to_hkl(h, k, l, wavelength, pos)
            self._verify_virtual_angles(h, k, l, pos, virtual_angles)
            if asdegrees:
                res_pos = Position.asdegrees(pos)
                res_virtual_angles = {
                    key: degrees(val) for key, val in virtual_angles.items()
                }
            else:
                res_pos = Position.asradians(pos)
                res_virtual_angles = copy(virtual_angles)
            results.append((res_pos, res_virtual_angles))

        return results

    def _calc_hkl_to_position(
        self, h: float, k: float, l: float, wavelength: float
    ) -> List[Tuple[Position, Dict[str, float]]]:
        if not self.constraints.is_fully_constrained():
            raise DiffcalcException(
                "Diffcalc is not fully constrained.\n"
                "Type 'help con' for instructions"
            )

        if not self.constraints.is_current_mode_implemented():
            raise DiffcalcException(
                "Sorry, the selected constraint combination is valid but "
                "is not implemented. Type 'help con' for implemented combinations"
            )

        # constraints are dictionaries
        ref_constraint = self.constraints._reference
        if ref_constraint:
            ref_constraint_name, ref_constraint_value = next(
                iter(ref_constraint.items())
            )
        det_constraint = self.constraints._detector
        naz_constraint = {"naz": self.constraints.naz} if self.constraints.naz else None
        samp_constraints = self.constraints._sample

        assert not (
            det_constraint and naz_constraint
        ), "Two 'detector' constraints given"

        h_phi = self.ubcalc.UB @ np.array([[h], [k], [l]])
        theta = (
            self.ubcalc.get_ttheta_from_hkl((h, k, l), 12.39842 / wavelength) / 2.0
        )  # __calc_theta(h_phi, wavelength)
        tau = angle_between_vectors(h_phi, self.ubcalc.n_phi)
        surf_tau = angle_between_vectors(h_phi, self.ubcalc.surf_nphi)

        if is_small(sin(tau)) and ref_constraint:
            if ref_constraint_name == "psi":
                raise DiffcalcException(
                    "Azimuthal angle 'psi' is undefined as reference and scattering vectors parallel.\n"
                    "Please constrain one of the sample angles or choose different reference vector orientation."
                )
            elif ref_constraint_name == "a_eq_b":
                raise DiffcalcException(
                    "Reference constraint 'a_eq_b' is redundant as reference and scattering vectors are parallel.\n"
                    "Please constrain one of the sample angles or choose different reference vector orientation."
                )
        if (
            is_small(sin(surf_tau))
            and ref_constraint
            and ref_constraint_name == "bin_eq_bout"
        ):
            raise DiffcalcException(
                "Reference constraint 'bin_eq_bout' is redundant as scattering vectors is parallel to the surface normal.\n"
                "Please select another constrain to define sample azimuthal orientation."
            )

        ### Reference constraint column ###

        n_phi = self.ubcalc.n_phi
        if ref_constraint:
            if {"psi", "a_eq_b", "alpha", "beta"}.issuperset(ref_constraint.keys()):
                # An angle for the reference vector (n) is given      (Section 5.2)
                alpha, _ = self._calc_remaining_reference_angles(
                    ref_constraint_name, ref_constraint_value, theta, tau
                )
            elif {"bin_eq_bout", "betain", "betaout"}.issuperset(ref_constraint.keys()):
                alpha, _ = self._calc_remaining_reference_angles(
                    ref_constraint_name, ref_constraint_value, theta, surf_tau
                )
                tau = surf_tau
                n_phi = self.ubcalc.surf_nphi

        solution_tuples = []
        if det_constraint or naz_constraint:

            if len(samp_constraints) == 1:
                for (
                    qaz,
                    naz,
                    delta,
                    nu,
                ) in self._calc_det_angles_given_det_or_naz_constraint(
                    det_constraint, naz_constraint, theta, tau, alpha
                ):
                    for (
                        mu,
                        eta,
                        chi,
                        phi,
                    ) in self._calc_sample_angles_from_one_sample_constraint(
                        samp_constraints, h_phi, theta, alpha, qaz, naz, n_phi
                    ):
                        solution_tuples.append((mu, delta, nu, eta, chi, phi))

            elif len(samp_constraints) == 2:
                if det_constraint:
                    det_constraint_name, det_constraint_val = next(
                        iter(det_constraint.items())
                    )
                    for delta, nu, qaz in self._calc_remaining_detector_angles(
                        det_constraint_name, det_constraint_val, theta
                    ):
                        for (
                            mu,
                            eta,
                            chi,
                            phi,
                        ) in self._calc_sample_angles_given_two_sample_and_detector(
                            samp_constraints, qaz, theta, h_phi, n_phi
                        ):
                            solution_tuples.append((mu, delta, nu, eta, chi, phi))

                else:
                    raise DiffcalcException(
                        "No code yet to handle this combination of detector and sample constraints!"
                    )

        elif len(samp_constraints) == 2:
            if ref_constraint_name == "psi":
                psi_vals = iter(
                    [
                        ref_constraint_value,
                    ]
                )
            else:
                psi_vals = self._calc_psi(alpha, theta, tau)
            for psi in psi_vals:
                solution_tuples.extend(
                    self._calc_sample_given_two_sample_and_reference(
                        samp_constraints, h_phi, theta, psi, n_phi
                    )
                )

        elif len(samp_constraints) == 3:
            solution_tuples.extend(
                self._calc_angles_given_three_sample_constraints(
                    samp_constraints,
                    h_phi,
                    theta,
                )
            )

        if not solution_tuples:
            raise DiffcalcException(
                "No solutions were found. "
                "Please consider using an alternative set of constraints."
            )

        tidy_solutions = [
            self._tidy_degenerate_solutions(Position(*pos, False))
            for pos in solution_tuples
        ]

        # def _find_duplicate_angles(el):
        #    idx, tpl = el
        #    for tmp_tpl in filtered_solutions[idx:]:
        #        if False not in [abs(x-y) < SMALL for x,y in zip(tmp_tpl, tpl)]:
        #            return False
        #    return True
        # merged_solution_tuples = filter(_find_duplicate_angles, enumerate(filtered_solutions, 1))
        position_pseudo_angles_pairs = self._create_position_pseudo_angles_pairs(
            tidy_solutions
        )
        if not position_pseudo_angles_pairs:
            raise DiffcalcException(
                "No solutions were found. Please consider using "
                "an alternative pseudo-angle constraints."
            )

        return position_pseudo_angles_pairs

    def _create_position_pseudo_angles_pairs(
        self, merged_solution_tuples: List[Position]
    ) -> List[Tuple[Position, Dict[str, float]]]:

        position_pseudo_angles_pairs = []
        for position in merged_solution_tuples:
            # Create position
            # position = self._tidy_degenerate_solutions(position)
            # if position.phi <= -pi + SMALL:
            #    position.phi += 2 * pi
            # pseudo angles calculated along the way were for the initial solution
            # and may be invalid for the chosen solution TODO: anglesToHkl need no
            # longer check the pseudo_angles as they will be generated with the
            # same function and it will prove nothing
            pseudo_angles = self.get_virtual_angles(position, False)
            try:
                for constraint in [
                    self.constraints._reference,
                    self.constraints._detector,
                ]:
                    for constraint_name, constraint_value in constraint.items():
                        if constraint_name == "a_eq_b":
                            assert radians_equivalent(
                                pseudo_angles["alpha"], pseudo_angles["beta"]
                            )
                        elif constraint_name == "bin_eq_bout":
                            assert radians_equivalent(
                                pseudo_angles["betain"], pseudo_angles["betaout"]
                            )
                        elif constraint_name not in pseudo_angles:
                            continue
                        else:
                            assert radians_equivalent(
                                constraint_value, pseudo_angles[constraint_name]
                            )
                position_pseudo_angles_pairs.append((position, pseudo_angles))
            except AssertionError:
                continue
        return position_pseudo_angles_pairs

    def _calc_N(self, Q: np.ndarray, n: np.ndarray) -> np.ndarray:
        """Return N as described by Equation 31."""
        Q = normalised(Q)
        n = normalised(n)
        if is_small(angle_between_vectors(Q, n)):
            # Replace the reference vector with an alternative vector from Eq.(78)
            def __key_func(v):
                return v[1]  # Workaround for mypy issue #9590

            idx_min, _ = min(
                enumerate([abs(Q[0, 0]), abs(Q[1, 0]), abs(Q[2, 0])]),
                key=__key_func,
            )
            idx_1, idx_2 = [idx for idx in range(3) if idx != idx_min]
            qval = sqrt(Q[idx_1, 0] * Q[idx_1, 0] + Q[idx_2, 0] * Q[idx_2, 0])
            n[idx_min, 0] = qval
            n[idx_1, 0] = -Q[idx_min, 0] * Q[idx_1, 0] / qval
            n[idx_2, 0] = -Q[idx_min, 0] * Q[idx_2, 0] / qval
            if is_small(norm(n)):
                n[idx_min, 0] = 0
                n[idx_1, 0] = Q[idx_2, 0] / qval
                n[idx_2, 0] = -Q[idx_1, 0] / qval
        Qxn = cross3(Q, n)
        QxnxQ = cross3(Qxn, Q)
        QxnxQ = normalised(QxnxQ)
        Qxn = normalised(Qxn)
        return np.array(
            [
                [Q[0, 0], QxnxQ[0, 0], Qxn[0, 0]],
                [Q[1, 0], QxnxQ[1, 0], Qxn[1, 0]],
                [Q[2, 0], QxnxQ[2, 0], Qxn[2, 0]],
            ]
        )

    def _calc_angle_between_naz_and_qaz(
        self, theta: float, alpha: float, tau: float
    ) -> float:
        # Equation 30:
        top = cos(tau) - sin(alpha) * sin(theta)
        bottom = cos(alpha) * cos(theta)
        if is_small(bottom):
            if is_small(cos(alpha)):
                raise ValueError("cos(alpha) is too small")
            if is_small(cos(theta)):
                raise ValueError("cos(theta) is too small")
        if is_small(sin(tau)):
            return 0.0
        return acos(bound(top / bottom))

    def _calc_psi(
        self,
        alpha: float,
        theta: float,
        tau: float,
        qaz: Optional[float] = None,
        naz: Optional[float] = None,
    ) -> Iterator[float]:
        """Calculate psi from Eq. (18), (25) and (28)."""
        sin_tau = sin(tau)
        cos_theta = cos(theta)
        if is_small(sin_tau):
            # The reference vector is parallel to the scattering vector
            yield float("nan")
        elif is_small(cos_theta):
            # Reflection is unreachable as theta angle is too close to 90 deg
            yield float("nan")
        elif is_small(sin(theta)):
            # Reflection is unreachable as |Q| is too small
            yield float("nan")
        else:
            cos_psi = (cos(tau) * sin(theta) - sin(alpha)) / cos_theta  # (28)
            if qaz is None or naz is None:
                try:
                    acos_psi = acos(bound(cos_psi / sin_tau))
                    if is_small(acos_psi):
                        yield 0.0
                    else:
                        for psi in [acos_psi, -acos_psi]:
                            yield psi
                except AssertionError:
                    print("WARNING: Diffcalc could not calculate an azimuth (psi)")
                    yield float("nan")
            else:
                sin_psi = cos(alpha) * sin(qaz - naz)
                sgn = sign(sin_tau)
                eps = sin_psi ** 2 + cos_psi ** 2
                sigma_ = eps / sin_tau ** 2 - 1
                if not is_small(sigma_):
                    print(
                        "WARNING: Diffcalc could not calculate a unique azimuth "
                        "(psi) because of loss of accuracy in numerical calculation"
                    )
                    yield float("nan")
                else:
                    psi = atan2(sgn * sin_psi, sgn * cos_psi)
                    yield psi

    def _calc_remaining_reference_angles(
        self, name: str, value: float, theta: float, tau: float
    ) -> Tuple[float, float]:
        """Return alpha and beta given one of a_eq_b, alpha, beta or psi."""
        UNREACHABLE_MSG = (
            "The current combination of constraints with %s = %.4f\n"
            "prohibits a solution for the specified reflection."
        )
        if name == "psi":
            psi = value
            # Equation 26 for alpha
            sin_alpha = cos(tau) * sin(theta) - cos(theta) * sin(tau) * cos(psi)
            if abs(sin_alpha) > 1 + SMALL:
                raise DiffcalcException(UNREACHABLE_MSG % (name, degrees(value)))
            alpha = asin(bound(sin_alpha))
            # Equation 27 for beta
            sin_beta = cos(tau) * sin(theta) + cos(theta) * sin(tau) * cos(psi)
            if abs(sin_beta) > 1 + SMALL:
                raise DiffcalcException(UNREACHABLE_MSG % (name, degrees(value)))

            beta = asin(bound(sin_beta))

        elif name == "a_eq_b" or name == "bin_eq_bout":
            alpha = beta = asin(cos(tau) * sin(theta))  # (24)

        elif name == "alpha" or name == "betain":
            alpha = value  # (24)
            sin_beta = 2 * sin(theta) * cos(tau) - sin(alpha)
            if abs(sin_beta) > 1 + SMALL:
                raise DiffcalcException(UNREACHABLE_MSG % (name, degrees(value)))
            beta = asin(sin_beta)

        elif name == "beta" or name == "betaout":
            beta = value
            sin_alpha = 2 * sin(theta) * cos(tau) - sin(beta)  # (24)
            if abs(sin_alpha) > 1 + SMALL:
                raise DiffcalcException(UNREACHABLE_MSG % (name, degrees(value)))

            alpha = asin(sin_alpha)

        return alpha, beta

    def _calc_det_angles_given_det_or_naz_constraint(
        self,
        det_constraint: Dict[str, Optional[float]],
        naz_constraint: Dict[str, Optional[float]],
        theta: float,
        tau: float,
        alpha: float,
    ) -> Iterator[Tuple[float, float, float, float]]:

        assert det_constraint or naz_constraint
        try:
            naz_qaz_angle = self._calc_angle_between_naz_and_qaz(theta, alpha, tau)
        except AssertionError:
            return
        if det_constraint:
            # One of the detector angles is given                 (Section 5.1)
            det_constraint_name, det_constraint_value = next(
                iter(det_constraint.items())
            )
            for delta, nu, qaz in self._calc_remaining_detector_angles(
                det_constraint_name, det_constraint_value, theta
            ):
                if is_small(naz_qaz_angle):
                    naz_angles = [
                        qaz,
                    ]
                else:
                    naz_angles = [qaz - naz_qaz_angle, qaz + naz_qaz_angle]
                for naz in naz_angles:
                    yield qaz, naz, delta, nu
        elif naz_constraint:  # The 'detector' angle naz is given:
            naz_name, naz = next(iter(naz_constraint.items()))
            assert naz_name == "naz"
            if is_small(naz_qaz_angle):
                qaz_angles = [
                    naz,
                ]
            else:
                qaz_angles = [naz - naz_qaz_angle, naz + naz_qaz_angle]
            for qaz in qaz_angles:
                for delta, nu, _ in self._calc_remaining_detector_angles(
                    "qaz", qaz, theta
                ):
                    yield qaz, naz, delta, nu

    def _calc_remaining_detector_angles(
        self, constraint_name: str, constraint_value: float, theta: float
    ) -> Iterator[Tuple[float, float, float]]:
        """Return delta, nu and qaz given one detector angle."""
        #                                                         (section 5.1)
        # Find qaz using various derivations of 17 and 18
        sin_2theta = sin(2 * theta)
        cos_2theta = cos(2 * theta)
        if is_small(sin_2theta):
            raise DiffcalcException(
                "No meaningful scattering vector (Q) can be found when "
                f"theta is so small {degrees(theta):.4f}."
            )

        if constraint_name == "delta":
            delta = constraint_value
            try:
                asin_qaz = asin(bound(sin(delta) / sin_2theta))  # (17 & 18)
            except AssertionError:
                return
            cos_delta = cos(delta)
            if is_small(cos_delta):
                # raise DiffcalcException(
                #    'The %s and %s circles are redundant when delta is constrained to %.0f degrees.'
                #    'Please change delta constraint or use 4-circle mode.' % ("nu", 'mu', delta * TODEG))
                print(
                    (
                        "DEGENERATE: with delta=90, %s is degenerate: choosing "
                        "%s = 0 (allowed because %s is unconstrained)"
                    )
                    % ("nu", "nu", "nu")
                )
                acos_nu = 1.0
            else:
                try:
                    acos_nu = acos(bound(cos_2theta / cos_delta))
                except AssertionError:
                    return
            if is_small(cos(asin_qaz)):
                qaz_angles = [
                    sign(asin_qaz) * pi / 2.0,
                ]
            else:
                qaz_angles = [asin_qaz, pi - asin_qaz]
            if is_small(acos_nu):
                nu_angles = [
                    0.0,
                ]
            else:
                nu_angles = [acos_nu, -acos_nu]
            for qaz, nu in product(qaz_angles, nu_angles):
                sgn_ref = sign(sin_2theta) * sign(cos(qaz))
                sgn_ratio = sign(sin(nu)) * sign(cos_delta)
                if sgn_ref == sgn_ratio:
                    yield delta, nu, qaz

        elif constraint_name == "nu":
            nu = constraint_value
            cos_nu = cos(nu)
            if is_small(cos_nu):
                raise DiffcalcException(
                    "The %s circle constraint to %.0f degrees is redundant."
                    "Please change this constraint or use 4-circle mode."
                    % ("nu", degrees(nu))
                )
            cos_delta = cos_2theta / cos(nu)
            cos_qaz = cos_delta * sin(nu) / sin_2theta
            try:
                acos_delta = acos(bound(cos_delta))
                acos_qaz = acos(bound(cos_qaz))
            except AssertionError:
                return
            if is_small(acos_qaz):
                qaz_angles = [
                    0.0,
                ]
            else:
                qaz_angles = [acos_qaz, -acos_qaz]
            if is_small(acos_delta):
                delta_angles = [
                    0.0,
                ]
            else:
                delta_angles = [acos_delta, -acos_delta]
            for qaz, delta in product(qaz_angles, delta_angles):
                sgn_ref = sign(sin(delta))
                sgn_ratio = sign(sin(qaz)) * sign(sin_2theta)
                if sgn_ref == sgn_ratio:
                    yield delta, nu, qaz

        elif constraint_name == "qaz":
            qaz = constraint_value
            asin_delta = asin(sin(qaz) * sin_2theta)
            if is_small(cos(asin_delta)):
                delta_angles = [
                    sign(asin_delta) * pi / 2.0,
                ]
            else:
                delta_angles = [asin_delta, pi - asin_delta]
            for delta in delta_angles:
                cos_delta = cos(delta)
                if is_small(cos_delta):
                    print(
                        (
                            "DEGENERATE: with delta=90, %s is degenerate: choosing "
                            "%s = 0 (allowed because %s is unconstrained)"
                        )
                        % ("nu", "nu", "nu")
                    )
                    # raise DiffcalcException(
                    #    'The %s circle is redundant when delta is at %.0f degrees.'
                    #    'Please change detector constraint or use 4-circle mode.' % ("nu", delta * TODEG))
                    nu = 0.0
                else:
                    sgn_delta = sign(cos_delta)
                    nu = atan2(
                        sgn_delta * sin_2theta * cos(qaz), sgn_delta * cos_2theta
                    )
                yield delta, nu, qaz
        else:
            raise DiffcalcException(
                constraint_name + " is not an explicit detector angle "
                "(naz cannot be handled here)"
            )

    def _calc_sample_angles_from_one_sample_constraint(
        self,
        samp_constraints: Dict[str, Optional[float]],
        h_phi: np.ndarray,
        theta: float,
        alpha: float,
        qaz: float,
        naz: float,
        n_phi: np.ndarray,
    ) -> Iterator[Tuple[float, float, float, float]]:

        sample_constraint_name, sample_value = next(iter(samp_constraints.items()))
        q_lab = np.array(
            [[cos(theta) * sin(qaz)], [-sin(theta)], [cos(theta) * cos(qaz)]]
        )  # (18)
        n_lab = np.array(
            [[cos(alpha) * sin(naz)], [-sin(alpha)], [cos(alpha) * cos(naz)]]
        )  # (20)
        yield from self._calc_remaining_sample_angles(
            sample_constraint_name, sample_value, q_lab, n_lab, h_phi, n_phi
        )

    def _calc_sample_given_two_sample_and_reference(
        self,
        samp_constraints: Dict[str, Optional[float]],
        h_phi: np.ndarray,
        theta: float,
        psi: float,
        n_phi: np.ndarray,
    ) -> Iterator[Tuple[float, float, float, float, float, float]]:

        for angles in self._calc_sample_angles_given_two_sample_and_reference(
            samp_constraints, psi, theta, h_phi, n_phi
        ):
            qaz, psi, mu, eta, chi, phi = angles
            values_in_deg = tuple(degrees(v) for v in angles)
            logger.debug(
                "Initial angles: xi=%.3f, psi=%.3f, mu=%.3f, "
                "eta=%.3f, chi=%.3f, phi=%.3f" % values_in_deg
            )  # Try to find a solution for each possible transformed xi

            logger.debug("")
            msg = "---Trying psi={:.3f}, qaz={:.3f}".format(degrees(psi), degrees(qaz))
            logger.debug(msg)

            for delta, nu, _ in self._calc_remaining_detector_angles("qaz", qaz, theta):
                logger.debug("delta=%.3f, %s=%.3f", degrees(delta), "nu", degrees(nu))
                # for mu, eta, chi, phi in self._generate_sample_solutions(
                #    mu, eta, chi, phi, samp_constraints.keys(), delta,
                #    nu, wavelength, (h, k, l), ref_constraint_name,
                #    ref_constraint_value):
                yield mu, delta, nu, eta, chi, phi

    def _calc_remaining_sample_angles(
        self,
        constraint_name: str,
        constraint_value: float,
        q_lab: np.ndarray,
        n_lab: np.ndarray,
        q_phi: np.ndarray,
        n_phi: np.ndarray,
    ) -> Iterator[Tuple[float, float, float, float]]:
        """Return phi, chi, eta and mu, given one of these."""
        #                                                         (section 5.3)

        N_lab = self._calc_N(q_lab, n_lab)
        N_phi = self._calc_N(q_phi, n_phi)
        Z = N_lab @ N_phi.T

        if constraint_name == "mu":  # (35)
            mu = constraint_value
            V = inv(rot_MU(mu)) @ N_lab @ N_phi.T
            try:
                acos_chi = acos(bound(V[2, 2]))
            except AssertionError:
                return
            if is_small(sin(acos_chi)):
                # chi ~= 0 or 180 and therefor phi || eta The solutions for phi
                # and eta here will be valid but will be chosen unpredictably.
                # Choose eta=0:
                #
                # tan(phi+eta)=v12/v11 from docs/extensions_to_yous_paper.wxm
                chi = acos_chi
                eta = 0.0
                phi = atan2(-V[1, 0], V[1, 1])
                logger.debug(
                    "Eta and phi cannot be chosen uniquely with chi so close "
                    "to 0 or 180. Returning phi=%.3f and eta=%.3f",
                    degrees(phi),
                    degrees(eta),
                )
                yield mu, eta, chi, phi
            else:
                for chi in [acos_chi, -acos_chi]:
                    sgn = sign(sin(chi))
                    phi = atan2(-sgn * V[2, 1], -sgn * V[2, 0])
                    eta = atan2(-sgn * V[1, 2], sgn * V[0, 2])
                    yield mu, eta, chi, phi

        elif constraint_name == "phi":  # (37)
            phi = constraint_value
            V = N_lab @ inv(N_phi) @ rot_PHI(phi).T
            try:
                asin_eta = asin(bound(V[0, 1]))
            except AssertionError:
                return
            if is_small(cos(asin_eta)):
                raise DiffcalcException(
                    "Chi and mu cannot be chosen uniquely "
                    "with eta so close to +/-90."
                )
            for eta in [asin_eta, pi - asin_eta]:
                sgn = sign(cos(eta))
                mu = atan2(sgn * V[2, 1], sgn * V[1, 1])
                chi = atan2(sgn * V[0, 2], sgn * V[0, 0])
                yield mu, eta, chi, phi

        elif constraint_name in ("eta", "chi"):
            if constraint_name == "eta":  # (39)
                eta = constraint_value
                cos_eta = cos(eta)
                if is_small(cos_eta):
                    # TODO: Not likely to happen in real world!?
                    raise DiffcalcException(
                        "Chi and mu cannot be chosen uniquely with eta "
                        "constrained so close to +-90."
                    )
                try:
                    asin_chi = asin(bound(Z[0, 2] / cos_eta))
                except AssertionError:
                    return
                all_eta = [
                    eta,
                ]
                all_chi = [asin_chi, pi - asin_chi]

            else:  # constraint_name == 'chi'                            # (40)
                chi = constraint_value
                sin_chi = sin(chi)
                if is_small(sin_chi):
                    raise DiffcalcException(
                        "Eta and phi cannot be chosen uniquely with chi "
                        "constrained so close to 0. (Please contact developer "
                        "if this case is useful for you)"
                    )
                try:
                    acos_eta = acos(bound(Z[0, 2] / sin_chi))
                except AssertionError:
                    return
                all_eta = [acos_eta, -acos_eta]
                all_chi = [
                    chi,
                ]

            for chi, eta in product(all_chi, all_eta):
                top_for_mu = Z[2, 2] * sin(eta) * sin(chi) + Z[1, 2] * cos(chi)
                bot_for_mu = -Z[2, 2] * cos(chi) + Z[1, 2] * sin(eta) * sin(chi)
                if is_small(top_for_mu) and is_small(bot_for_mu):
                    # chi == +-90, eta == 0/180 and therefore phi || mu cos(chi) ==
                    # 0 and sin(eta) == 0 Experience shows that even though e.g.
                    # the z[2, 2] and z[1, 2] values used to calculate mu may be
                    # basically 0 (1e-34) their ratio in every case tested so far
                    # still remains valid and using them will result in a phi
                    # solution that is continuous with neighbouring positions.
                    #
                    # We cannot test phi minus mu here unfortunately as the final
                    # phi and mu solutions have not yet been chosen (they may be
                    # +-x or 180+-x). Otherwise we could choose a sensible solution
                    # here if the one found was incorrect.

                    # tan(phi+eta)=v12/v11 from extensions_to_yous_paper.wxm
                    # phi_minus_mu = -atan2(Z[2, 0], Z[1, 1])
                    raise DiffcalcException(
                        "Mu cannot be chosen uniquely as mu || phi with chi so close "
                        "to +/-90 and eta so close 0 or 180.\nPlease choose "
                        "a different set of constraints."
                    )
                mu = atan2(-top_for_mu, -bot_for_mu)  # (41)

                top_for_phi = Z[0, 1] * cos(eta) * cos(chi) - Z[0, 0] * sin(eta)
                bot_for_phi = Z[0, 1] * sin(eta) + Z[0, 0] * cos(eta) * cos(chi)
                if is_small(bot_for_phi) and is_small(top_for_phi):
                    DiffcalcException(
                        "Phi cannot be chosen uniquely as mu || phi with chi so close "
                        "to +/-90 and eta so close 0 or 180.\nPlease choose a "
                        "different set of constraints."
                    )
                phi = atan2(top_for_phi, bot_for_phi)  # (42)
                yield mu, eta, chi, phi

        else:
            raise DiffcalcException("Given angle must be one of phi, chi, eta or mu")

    def _calc_angles_given_three_sample_constraints(
        self,
        samp_constraints: Dict[str, Optional[float]],
        h_phi: np.ndarray,
        theta: float,
    ) -> Iterator[Tuple[float, float, float, float, float, float]]:
        def __get_last_sample_angle(A: float, B: float, C: float) -> List[float]:
            if is_small(A) and is_small(B):
                raise DiffcalcException(
                    "Sample orientation cannot be chosen uniquely. Please choose a different set of constraints."
                )
            ks = atan2(A, B)
            acos_alp = acos(bound(C / sqrt(A ** 2 + B ** 2)))
            if is_small(acos_alp):
                alp_list = [
                    ks,
                ]
            else:
                alp_list = [acos_alp + ks, -acos_alp + ks]
            return alp_list

        def __get_qaz_value(mu: float, eta: float, chi: float, phi: float) -> float:
            V0 = (
                h2 * cos(eta) * sin(chi)
                + (h0 * cos(chi) * cos(eta) + h1 * sin(eta)) * cos(phi)
                + (h1 * cos(chi) * cos(eta) - h0 * sin(eta)) * sin(phi)
            )
            V2 = (
                -h2 * sin(chi) * sin(eta) * sin(mu)
                + h2 * cos(chi) * cos(mu)
                - (
                    h0 * cos(mu) * sin(chi)
                    + (h0 * cos(chi) * sin(eta) - h1 * cos(eta)) * sin(mu)
                )
                * cos(phi)
                - (
                    h1 * cos(mu) * sin(chi)
                    + (h1 * cos(chi) * sin(eta) + h0 * cos(eta)) * sin(mu)
                )
                * sin(phi)
            )
            sgn_theta = sign(cos(theta))
            qaz = atan2(sgn_theta * V0, sgn_theta * V2)
            return qaz

        h_phi_norm = normalised(h_phi)  # (68,69)
        h0, h1, h2 = h_phi_norm[0, 0], h_phi_norm[1, 0], h_phi_norm[2, 0]

        if "mu" not in samp_constraints:
            eta = self.constraints._sample["eta"]
            chi = self.constraints._sample["chi"]
            phi = self.constraints._sample["phi"]

            A = h0 * cos(phi) * sin(chi) + h1 * sin(chi) * sin(phi) - h2 * cos(chi)
            B = (
                -h2 * sin(chi) * sin(eta)
                - (h0 * cos(chi) * sin(eta) - h1 * cos(eta)) * cos(phi)
                - (h1 * cos(chi) * sin(eta) + h0 * cos(eta)) * sin(phi)
            )
            C = -sin(theta)
            try:
                mu_vals = __get_last_sample_angle(A, B, C)
            except AssertionError:
                return
            for mu in mu_vals:
                qaz = __get_qaz_value(mu, eta, chi, phi)
                logger.debug("--- Trying mu:%.f qaz_%.f", degrees(mu), degrees(qaz))
                for delta, nu, _ in self._calc_remaining_detector_angles(
                    "qaz", qaz, theta
                ):
                    logger.debug(
                        "delta=%.3f, %s=%.3f", degrees(delta), "nu", degrees(nu)
                    )
                    yield mu, delta, nu, eta, chi, phi

        elif "eta" not in samp_constraints:
            mu = self.constraints._sample["mu"]
            chi = self.constraints._sample["chi"]
            phi = self.constraints._sample["phi"]

            A = (
                -h0 * cos(chi) * cos(mu) * cos(phi)
                - h1 * cos(chi) * cos(mu) * sin(phi)
                - h2 * cos(mu) * sin(chi)
            )
            B = h1 * cos(mu) * cos(phi) - h0 * cos(mu) * sin(phi)
            C = (
                -h0 * cos(phi) * sin(chi) * sin(mu)
                - h1 * sin(chi) * sin(mu) * sin(phi)
                + h2 * cos(chi) * sin(mu)
                - sin(theta)
            )
            try:
                eta_vals = __get_last_sample_angle(A, B, C)
            except AssertionError:
                return
            for eta in eta_vals:
                qaz = __get_qaz_value(mu, eta, chi, phi)
                logger.debug("--- Trying eta:%.f qaz_%.f", degrees(eta), degrees(qaz))
                for delta, nu, _ in self._calc_remaining_detector_angles(
                    "qaz", qaz, theta
                ):
                    logger.debug(
                        "delta=%.3f, %s=%.3f", degrees(delta), "nu", degrees(nu)
                    )
                    yield mu, delta, nu, eta, chi, phi

        elif "chi" not in samp_constraints:
            mu = self.constraints._sample["mu"]
            eta = self.constraints._sample["eta"]
            phi = self.constraints._sample["phi"]

            A = (
                -h2 * cos(mu) * sin(eta)
                + h0 * cos(phi) * sin(mu)
                + h1 * sin(mu) * sin(phi)
            )
            B = (
                -h0 * cos(mu) * cos(phi) * sin(eta)
                - h1 * cos(mu) * sin(eta) * sin(phi)
                - h2 * sin(mu)
            )
            C = (
                -h1 * cos(eta) * cos(mu) * cos(phi)
                + h0 * cos(eta) * cos(mu) * sin(phi)
                - sin(theta)
            )
            try:
                chi_vals = __get_last_sample_angle(A, B, C)
            except AssertionError:
                return
            for chi in chi_vals:
                qaz = __get_qaz_value(mu, eta, chi, phi)
                logger.debug("--- Trying chi:%.f qaz_%.f", degrees(chi), degrees(qaz))
                for delta, nu, _ in self._calc_remaining_detector_angles(
                    "qaz", qaz, theta
                ):
                    logger.debug("delta=%.3f, nu=%.3f", degrees(delta), degrees(nu))
                    yield mu, delta, nu, eta, chi, phi

        elif "phi" not in samp_constraints:
            mu = self.constraints._sample["mu"]
            eta = self.constraints._sample["eta"]
            chi = self.constraints._sample["chi"]

            A = h1 * sin(chi) * sin(mu) - (
                h1 * cos(chi) * sin(eta) + h0 * cos(eta)
            ) * cos(mu)
            B = h0 * sin(chi) * sin(mu) - (
                h0 * cos(chi) * sin(eta) - h1 * cos(eta)
            ) * cos(mu)
            C = (
                h2 * cos(mu) * sin(chi) * sin(eta)
                + h2 * cos(chi) * sin(mu)
                - sin(theta)
            )
            try:
                phi_vals = __get_last_sample_angle(A, B, C)
            except AssertionError:
                return
            for phi in phi_vals:
                qaz = __get_qaz_value(mu, eta, chi, phi)
                logger.debug("--- Trying phi:%.f qaz_%.f", degrees(phi), degrees(qaz))
                for delta, nu, _ in self._calc_remaining_detector_angles(
                    "qaz", qaz, theta
                ):
                    logger.debug("delta=%.3f, nu=%.3f", degrees(delta), degrees(nu))
                    yield mu, delta, nu, eta, chi, phi
        else:
            raise DiffcalcException(
                "Internal error: Invalid set of sample constraints."
            )

    def _calc_sample_angles_given_two_sample_and_reference(
        self,
        samp_constraints: Dict[str, Optional[float]],
        psi: float,
        theta: float,
        q_phi: np.ndarray,
        n_phi: np.ndarray,
    ) -> Iterator[Tuple[float, float, float, float, float, float]]:
        """Return sample angles.

        Available combinations:
        chi, phi, reference
        mu, eta, reference,
        chi, eta, reference
        chi, mu, reference
        mu, phi, reference
        eta, phi, reference
        """

        def __get_phi_and_qaz(chi: float, eta: float, mu: float) -> Tuple[float, float]:
            a = sin(chi) * cos(eta)
            b = sin(chi) * sin(eta) * sin(mu) - cos(chi) * cos(mu)
            sin_qaz = V[2, 0] * a - V[2, 2] * b
            cos_qaz = -V[2, 2] * a - V[2, 0] * b
            # atan2_xi = atan2(V[2, 2] * a + V[2, 0] * b,
            #           V[2, 0] * a - V[2, 2] * b)                        # (54)
            qaz = atan2(sin_qaz, cos_qaz)  # (54)

            a = sin(chi) * sin(mu) - cos(mu) * cos(chi) * sin(eta)
            b = cos(mu) * cos(eta)
            phi = atan2(V[1, 1] * a - V[0, 1] * b, V[0, 1] * a + V[1, 1] * b)  # (55)
            #        if is_small(mu+pi/2) and is_small(eta) and False:
            #            phi_general = phi
            #            # solved in extensions_to_yous_paper.wxm
            #            phi = atan2(V[1, 1], V[0, 1])
            #            logger.debug("phi = %.3f or %.3f (std)",
            #                        phi*TODEG, phi_general*TODEG )

            return qaz, phi

        def __get_chi_and_qaz(mu: float, eta: float) -> Tuple[float, float]:
            A = sin(mu)
            B = -cos(mu) * sin(eta)
            sin_chi = A * V[1, 0] + B * V[1, 2]
            cos_chi = B * V[1, 0] - A * V[1, 2]
            if is_small(sin_chi) and is_small(cos_chi):
                raise DiffcalcException(
                    "Chi cannot be chosen uniquely. Please choose a different set of constraints."
                )
            chi = atan2(sin_chi, cos_chi)

            A = sin(eta)
            B = cos(eta) * sin(mu)
            sin_qaz = A * V[0, 1] + B * V[2, 1]
            cos_qaz = B * V[0, 1] - A * V[2, 1]
            qaz = atan2(sin_qaz, cos_qaz)
            return qaz, chi

        N_phi = self._calc_N(q_phi, n_phi)
        THETA = z_rotation(-theta)
        PSI = x_rotation(psi)

        if "chi" in samp_constraints and "phi" in samp_constraints:

            chi = samp_constraints["chi"]
            phi = samp_constraints["phi"]

            CHI = rot_CHI(chi)
            PHI = rot_PHI(phi)
            V = CHI @ PHI @ N_phi @ PSI.T @ THETA.T  # (46)

            # atan2_xi = atan2(-V[2, 0], V[2, 2])
            # atan2_eta = atan2(-V[0, 1], V[1, 1])
            # atan2_mu = atan2(-V[2, 1], sqrt(V[2, 2] ** 2 + V[2, 0] ** 2))
            try:
                asin_mu = asin(bound(-V[2, 1]))
            except AssertionError:
                return
            for mu in [asin_mu, pi - asin_mu]:
                sgn_cosmu = sign(cos(mu))
                # xi = atan2(-sgn_cosmu * V[2, 0], sgn_cosmu * V[2, 2])
                qaz = atan2(
                    sgn_cosmu * V[2, 2],
                    sgn_cosmu * V[2, 0],
                )
                eta = atan2(-sgn_cosmu * V[0, 1], sgn_cosmu * V[1, 1])
                yield qaz, psi, mu, eta, chi, phi

        elif "mu" in samp_constraints and "eta" in samp_constraints:

            mu = samp_constraints["mu"]
            eta = samp_constraints["eta"]

            V = N_phi @ PSI.T @ THETA.T  # (49)
            try:
                bot = bound(
                    -V[2, 1] / sqrt(sin(eta) ** 2 * cos(mu) ** 2 + sin(mu) ** 2)
                )
            except AssertionError:
                return
            if is_small(cos(mu) * sin(eta)):
                eps = atan2(sin(eta) * cos(mu), sin(mu))
                chi_vals = [eps + acos(bot), eps - acos(bot)]
            else:
                eps = atan2(sin(mu), sin(eta) * cos(mu))
                chi_vals = [asin(bot) - eps, pi - asin(bot) - eps]  # (52)

            ## Choose final chi solution here to obtain compatable xi and mu
            ## TODO: This temporary solution works only for one case used on i07
            ##       Return a list of possible solutions?
            # if is_small(eta) and is_small(mu + pi / 2):
            #    for chi in _generate_transformed_values(chi_orig):
            #        if  pi / 2 <= chi < pi:
            #            break
            # else:
            #    chi = chi_orig

            for chi in chi_vals:
                qaz, phi = __get_phi_and_qaz(chi, eta, mu)
                yield qaz, psi, mu, eta, chi, phi

        elif "chi" in samp_constraints and "eta" in samp_constraints:

            chi = samp_constraints["chi"]
            eta = samp_constraints["eta"]

            V = N_phi @ PSI.T @ THETA.T  # (49)
            try:
                bot = bound(
                    -V[2, 1] / sqrt(sin(eta) ** 2 * sin(chi) ** 2 + cos(chi) ** 2)
                )
            except AssertionError:
                return
            if is_small(cos(chi)):
                eps = atan2(cos(chi), sin(chi) * sin(eta))
                mu_vals = [eps + acos(bot), eps - acos(bot)]
            else:
                eps = atan2(sin(chi) * sin(eta), cos(chi))
                mu_vals = [asin(bot) - eps, pi - asin(bot) - eps]  # (52)

            for mu in mu_vals:
                qaz, phi = __get_phi_and_qaz(chi, eta, mu)
                yield qaz, psi, mu, eta, chi, phi

        elif "chi" in samp_constraints and "mu" in samp_constraints:

            chi = samp_constraints["chi"]
            mu = samp_constraints["mu"]

            V = N_phi @ PSI.T @ THETA.T  # (49)

            try:
                asin_eta = asin(
                    bound((-V[2, 1] - cos(chi) * sin(mu)) / (sin(chi) * cos(mu)))
                )
            except AssertionError:
                return

            for eta in [asin_eta, pi - asin_eta]:
                qaz, phi = __get_phi_and_qaz(chi, eta, mu)
                yield qaz, psi, mu, eta, chi, phi

        elif "mu" in samp_constraints and "phi" in samp_constraints:

            mu = samp_constraints["mu"]
            phi = samp_constraints["phi"]

            PHI = rot_PHI(phi)
            V = THETA @ PSI @ inv(N_phi) @ PHI.T

            if is_small(cos(mu)):
                raise DiffcalcException(
                    "Eta cannot be chosen uniquely. Please choose a different set of constraints."
                )
            try:
                acos_eta = acos(bound(V[1, 1] / cos(mu)))
            except AssertionError:
                return
            for eta in [acos_eta, -acos_eta]:
                qaz, chi = __get_chi_and_qaz(mu, eta)
                yield qaz, psi, mu, eta, chi, phi

        elif "eta" in samp_constraints and "phi" in samp_constraints:

            eta = samp_constraints["eta"]
            phi = samp_constraints["phi"]

            PHI = rot_PHI(phi)
            V = THETA @ PSI @ inv(N_phi) @ PHI.T

            if is_small(cos(eta)):
                raise DiffcalcException(
                    "Mu cannot be chosen uniquely. Please choose a different set of constraints."
                )
            try:
                acos_mu = acos(bound(V[1, 1] / cos(eta)))
            except AssertionError:
                return
            for mu in [acos_mu, -acos_mu]:
                qaz, chi = __get_chi_and_qaz(mu, eta)
                yield qaz, psi, mu, eta, chi, phi

        else:
            raise DiffcalcException(
                "No code yet to handle this combination of 2 sample "
                "constraints and one reference!:" + str(samp_constraints)
            )

    def _calc_sample_angles_given_two_sample_and_detector(
        self,
        samp_constraints: Dict[str, Optional[float]],
        qaz: float,
        theta: float,
        q_phi: np.ndarray,
        n_phi: np.ndarray,
    ) -> Iterator[Tuple[float, float, float, float]]:
        """Return sample angles.

        Available combinations:
        chi, phi, detector
        mu, eta, detector
        mu, phi, detector
        mu, chi, detector
        eta, phi, detector
        eta, chi, detector
        """
        N_phi = self._calc_N(q_phi, n_phi)

        if (
            ("mu" in samp_constraints and "eta" in samp_constraints)
            or ("omega" in samp_constraints and "bisect" in samp_constraints)
            or ("mu" in samp_constraints and "bisect" in samp_constraints)
            or ("eta" in samp_constraints and "bisect" in samp_constraints)
        ):

            if "mu" in samp_constraints and "eta" in samp_constraints:
                mu_vals = [
                    samp_constraints["mu"],
                ]
                eta_vals = [
                    samp_constraints["eta"],
                ]
            elif "omega" in samp_constraints and "bisect" in samp_constraints:
                omega = samp_constraints["omega"]
                atan_mu = atan(tan(theta + omega) * cos(qaz))
                asin_eta = asin(sin(theta + omega) * sin(qaz))
                mu_vals = [atan_mu, atan_mu + pi]
                if is_small(abs(asin_eta) - pi / 2):
                    eta_vals = [
                        sign(asin_eta) * pi / 2,
                    ]
                else:
                    eta_vals = [asin_eta, pi - asin_eta]

            elif "mu" in samp_constraints and "bisect" in samp_constraints:
                mu_vals = [
                    samp_constraints["mu"],
                ]
                cos_qaz = cos(qaz)
                tan_mu = tan(samp_constraints["mu"])
                # Vertical scattering geometry with omega = 0
                if is_small(cos_qaz):
                    if is_small(tan_mu):
                        thomega_vals = [
                            theta,
                        ]
                    else:
                        return
                else:
                    atan_thomega = atan(tan_mu / cos_qaz)
                    thomega_vals = [atan_thomega, pi + atan_thomega]
                eta_vals = []
                for thomega in thomega_vals:
                    asin_eta = asin(sin(thomega) * sin(qaz))
                    if is_small(abs(asin_eta) - pi / 2):
                        eta_vals.extend(
                            [
                                sign(asin_eta) * pi / 2,
                            ]
                        )
                    else:
                        eta_vals.extend([asin_eta, pi - asin_eta])

            elif "eta" in samp_constraints and "bisect" in samp_constraints:
                eta_vals = [
                    samp_constraints["eta"],
                ]
                sin_qaz = sin(qaz)
                sin_eta = sin(samp_constraints["eta"])
                # Horizontal scattering geometry with omega = 0
                if is_small(sin_qaz):
                    if is_small(sin_eta):
                        thomega_vals = [
                            theta,
                        ]
                    else:
                        return
                else:
                    asin_thomega = asin(sin_eta / sin_qaz)
                    if is_small(abs(asin_thomega) - pi / 2):
                        thomega_vals = [
                            sign(asin_thomega) * pi / 2,
                        ]
                    else:
                        thomega_vals = [asin_thomega, pi - asin_thomega]
                mu_vals = []
                for thomega in thomega_vals:
                    atan_mu = atan(tan(thomega) * cos(qaz))
                    mu_vals.extend([atan_mu, pi + atan_mu])

            for mu, eta in product(mu_vals, eta_vals):
                F = y_rotation(qaz - pi / 2.0)
                THETA = z_rotation(-theta)
                V = rot_ETA(eta).T @ rot_MU(mu).T @ F @ THETA  # (56)

                phi_vals = []
                try:
                    # For the case of (00l) reflection, where N_phi[0,0] = N_phi[1,0] = 0
                    if is_small(N_phi[0, 0]) and is_small(N_phi[1, 0]):
                        raise DiffcalcException(
                            "Phi cannot be chosen uniquely as q || phi and no reference "
                            "vector or phi constraints have been set.\nPlease choose a different "
                            "set of constraints."
                        )
                    bot = bound(-V[1, 0] / sqrt(N_phi[0, 0] ** 2 + N_phi[1, 0] ** 2))
                    eps = atan2(N_phi[1, 0], N_phi[0, 0])
                    phi_vals = [asin(bot) + eps, pi - asin(bot) + eps]  # (59)
                except AssertionError:
                    continue
                for phi in phi_vals:
                    a = N_phi[0, 0] * cos(phi) + N_phi[1, 0] * sin(phi)
                    chi = atan2(
                        N_phi[2, 0] * V[0, 0] - a * V[2, 0],
                        N_phi[2, 0] * V[2, 0] + a * V[0, 0],
                    )  # (60)
                    yield mu, eta, chi, phi

        elif "chi" in samp_constraints and "phi" in samp_constraints:

            chi = samp_constraints["chi"]
            phi = samp_constraints["phi"]

            CHI = rot_CHI(chi)
            PHI = rot_PHI(phi)
            V = CHI @ PHI @ N_phi  # (62)

            try:
                bot = bound(
                    V[2, 0] / sqrt(cos(qaz) ** 2 * cos(theta) ** 2 + sin(theta) ** 2)
                )
            except AssertionError:
                return
            eps = atan2(-cos(qaz) * cos(theta), sin(theta))
            for mu in [asin(bot) + eps, pi - asin(bot) + eps]:
                a = cos(theta) * sin(qaz)
                b = -cos(theta) * sin(mu) * cos(qaz) + cos(mu) * sin(theta)
                X = V[1, 0] * a + V[0, 0] * b
                Y = V[0, 0] * a - V[1, 0] * b
                if is_small(X) and is_small(Y):
                    raise DiffcalcException(
                        "Eta cannot be chosen uniquely as q || eta and no reference "
                        "vector or eta constraints have been set.\nPlease choose a different "
                        "set of constraints."
                    )
                eta = atan2(X, Y)

                # a = -cos(mu) * cos(qaz) * sin(theta) + sin(mu) * cos(theta)
                # b = cos(mu) * sin(qaz)
                # psi = atan2(-V[2, 2] * a - V[2, 1] * b, V[2, 1] * a - V[2, 2] * b)
                yield mu, eta, chi, phi

        elif "mu" in samp_constraints and "phi" in samp_constraints:

            mu = samp_constraints["mu"]
            phi = samp_constraints["phi"]

            F = y_rotation(qaz - pi / 2.0)
            THETA = z_rotation(-theta)
            V = rot_MU(mu).T @ F @ THETA
            E = rot_PHI(phi) @ N_phi

            try:
                bot = bound(-V[2, 0] / sqrt(E[0, 0] ** 2 + E[2, 0] ** 2))
            except AssertionError:
                return
            eps = atan2(E[2, 0], E[0, 0])
            for chi in [asin(bot) + eps, pi - asin(bot) + eps]:
                a = E[0, 0] * cos(chi) + E[2, 0] * sin(chi)
                eta = atan2(
                    V[0, 0] * E[1, 0] - V[1, 0] * a, V[0, 0] * a + V[1, 0] * E[1, 0]
                )
                yield mu, eta, chi, phi

        elif "mu" in samp_constraints and "chi" in samp_constraints:

            mu = samp_constraints["mu"]
            chi = samp_constraints["chi"]

            V20 = cos(mu) * cos(qaz) * cos(theta) + sin(mu) * sin(theta)
            A = N_phi[1, 0]
            B = N_phi[0, 0]
            if is_small(sin(chi)):
                raise DiffcalcException(
                    "Degenerate configuration with phi || eta axes cannot be set uniquely. Please choose a different set of constraints."
                )
            if is_small(A) and is_small(B):
                raise DiffcalcException(
                    "Phi cannot be chosen uniquely. Please choose a different set of constraints."
                )
            else:
                ks = atan2(A, B)
            try:
                acos_phi = acos(
                    bound(
                        (N_phi[2, 0] * cos(chi) - V20)
                        / (sin(chi) * sqrt(A ** 2 + B ** 2))
                    )
                )
            except AssertionError:
                return
            if is_small(acos_phi):
                phi_list = [
                    ks,
                ]
            else:
                phi_list = [acos_phi + ks, -acos_phi + ks]
            for phi in phi_list:
                A00 = -cos(qaz) * cos(theta) * sin(mu) + cos(mu) * sin(theta)
                B00 = sin(qaz) * cos(theta)
                V00 = (
                    N_phi[0, 0] * cos(chi) * cos(phi)
                    + N_phi[1, 0] * cos(chi) * sin(phi)
                    + N_phi[2, 0] * sin(chi)
                )
                V10 = N_phi[1, 0] * cos(phi) - N_phi[0, 0] * sin(phi)
                sin_eta = V00 * A00 + V10 * B00
                cos_eta = V00 * B00 - V10 * A00
                if is_small(A00) and is_small(B00):
                    raise DiffcalcException(
                        "Eta cannot be chosen uniquely. Please choose a different set of constraints."
                    )
                eta = atan2(sin_eta, cos_eta)
                yield mu, eta, chi, phi

        elif "eta" in samp_constraints and "phi" in samp_constraints:

            eta = samp_constraints["eta"]
            phi = samp_constraints["phi"]

            X = N_phi[2, 0]
            Y = N_phi[0, 0] * cos(phi) + N_phi[1, 0] * sin(phi)
            if is_small(X) and is_small(Y):
                raise DiffcalcException(
                    "Chi cannot be chosen uniquely as q || chi and no reference "
                    "vector or chi constraints have been set.\nPlease choose a different "
                    "set of constraints."
                )

            V = (N_phi[1, 0] * cos(phi) - N_phi[0, 0] * sin(phi)) * tan(eta)
            sgn = sign(cos(eta))
            eps = atan2(X * sgn, Y * sgn)
            try:
                acos_rhs = acos(
                    bound(
                        (sin(qaz) * cos(theta) / cos(eta) - V) / sqrt(X ** 2 + Y ** 2)
                    )
                )
            except AssertionError:
                return
            if is_small(acos_rhs):
                acos_list = [
                    eps,
                ]
            else:
                acos_list = [eps + acos_rhs, eps - acos_rhs]
            for chi in acos_list:
                A = (N_phi[0, 0] * cos(phi) + N_phi[1, 0] * sin(phi)) * sin(
                    chi
                ) - N_phi[2, 0] * cos(chi)
                B = (
                    -N_phi[2, 0] * sin(chi) * sin(eta)
                    - cos(chi)
                    * sin(eta)
                    * (N_phi[0, 0] * cos(phi) + N_phi[1, 0] * sin(phi))
                    - cos(eta) * (N_phi[0, 0] * sin(phi) - N_phi[1, 0] * cos(phi))
                )
                ks = atan2(A, B)
                mu = atan2(cos(theta) * cos(qaz), -sin(theta)) + ks
                yield mu, eta, chi, phi

        elif "eta" in samp_constraints and "chi" in samp_constraints:

            eta = samp_constraints["eta"]
            chi = samp_constraints["chi"]

            A = N_phi[1, 0] * cos(chi) * cos(eta) - N_phi[0, 0] * sin(eta)
            B = N_phi[0, 0] * cos(chi) * cos(eta) + N_phi[1, 0] * sin(eta)
            if is_small(A) and is_small(B):
                raise DiffcalcException(
                    "Phi cannot be chosen uniquely. Please choose a different set of constraints."
                )
            else:
                ks = atan2(A, B)
            try:
                acos_V00 = acos(
                    bound(
                        (cos(theta) * sin(qaz) - N_phi[2, 0] * cos(eta) * sin(chi))
                        / sqrt(A ** 2 + B ** 2)
                    )
                )
            except AssertionError:
                return
            if is_small(acos_V00):
                phi_list = [
                    ks,
                ]
            else:
                phi_list = [acos_V00 + ks, -acos_V00 + ks]
            for phi in phi_list:
                A10 = (
                    N_phi[0, 0] * cos(phi) * sin(chi)
                    + N_phi[1, 0] * sin(chi) * sin(phi)
                    - N_phi[2, 0] * cos(chi)
                )
                B10 = (
                    -N_phi[2, 0] * sin(chi) * sin(eta)
                    - (cos(chi) * cos(phi) * sin(eta) + cos(eta) * sin(phi))
                    * N_phi[0, 0]
                    - (cos(chi) * sin(eta) * sin(phi) - cos(eta) * cos(phi))
                    * N_phi[1, 0]
                )
                V10 = -sin(theta)
                A20 = (
                    -N_phi[2, 0] * sin(chi) * sin(eta)
                    - (cos(chi) * cos(phi) * sin(eta) + cos(eta) * sin(phi))
                    * N_phi[0, 0]
                    - (cos(chi) * sin(eta) * sin(phi) - cos(eta) * cos(phi))
                    * N_phi[1, 0]
                )
                B20 = (
                    -N_phi[0, 0] * cos(phi) * sin(chi)
                    - N_phi[1, 0] * sin(chi) * sin(phi)
                    + N_phi[2, 0] * cos(chi)
                )
                V20 = cos(qaz) * cos(theta)
                sin_mu = (V10 * B20 - V20 * B10) * sign(A10 * B20 - A20 * B10)
                cos_mu = (V10 * A20 - V20 * A10) * sign(B10 * A20 - B20 * A10)
                if is_small(sin_mu) and is_small(cos_mu):
                    raise DiffcalcException(
                        "Mu cannot be chosen uniquely. Please choose a different set of constraints."
                    )
                mu = atan2(sin_mu, cos_mu)
                yield mu, eta, chi, phi

        else:
            raise DiffcalcException(
                "No code yet to handle this combination of 2 sample "
                "constraints and one detector!:" + str(samp_constraints)
            )

    def _tidy_degenerate_solutions(
        self, pos: Position, print_degenerate: bool = False
    ) -> Position:

        detector_like_constraint = self.constraints._detector or self.constraints.naz
        nu_constrained_to_0 = is_small(pos.nu) and detector_like_constraint
        mu_constrained_to_0 = is_small(pos.mu) and "mu" in self.constraints._sample
        delta_constrained_to_0 = is_small(pos.delta) and detector_like_constraint
        eta_constrained_to_0 = is_small(pos.eta) and "eta" in self.constraints._sample
        phi_not_constrained = "phi" not in self.constraints._sample

        if (
            nu_constrained_to_0
            and mu_constrained_to_0
            and phi_not_constrained
            and is_small(pos.chi)
        ):
            # constrained to vertical 4-circle like mode
            # phi || eta
            desired_eta = pos.delta / 2.0
            eta_diff = desired_eta - pos.eta
            if print_degenerate:
                print(
                    "DEGENERATE: with chi=0, phi and eta are colinear:"
                    "choosing eta = delta/2 by adding % 7.3f to eta and "
                    "removing it from phi. (mu=nu=0 only)" % degrees(eta_diff)
                )
                print("            original:", pos)
            newpos = Position(
                pos.mu,
                pos.delta,
                pos.nu,
                desired_eta,
                pos.chi,
                pos.phi - eta_diff,
                False,
            )

        elif (
            delta_constrained_to_0
            and eta_constrained_to_0
            and phi_not_constrained
            and is_small(pos.chi - pi / 2)
        ):
            # constrained to horizontal 4-circle like mode
            # phi || mu
            desired_mu = pos.nu / 2.0
            mu_diff = desired_mu - pos.mu
            if print_degenerate:
                print(
                    "DEGENERATE: with chi=90, phi and mu are colinear: choosing"
                    " mu = nu/2 by adding % 7.3f to mu and to phi. "
                    "(delta=eta=0 only)" % degrees(mu_diff)
                )
                print("            original:", pos)
            newpos = Position(
                desired_mu,
                pos.delta,
                pos.nu,
                pos.eta,
                pos.chi,
                pos.phi + mu_diff,
                False,
            )
        else:
            newpos = pos
        return newpos

    def _theta_and_qaz_from_detector_angles(
        self, delta: float, nu: float
    ) -> Tuple[float, float]:
        # Equation 19:
        cos_2theta = cos(delta) * cos(nu)
        theta = acos(cos_2theta) / 2.0
        sgn = sign(sin(2.0 * theta))
        qaz = atan2(sgn * sin(delta), sgn * cos(delta) * sin(nu))
        return theta, qaz

    def _verify_pos_map_to_hkl(
        self, h: float, k: float, l: float, wavelength: float, pos: Position
    ) -> None:
        hkl = self.get_hkl(pos, wavelength)
        e = 0.001
        if (abs(hkl[0] - h) > e) or (abs(hkl[1] - k) > e) or (abs(hkl[2] - l) > e):
            s = "ERROR: The angles calculated for hkl=({:f},{:f},{:f}) were {}.\n".format(
                h,
                k,
                l,
                str(pos),
            )
            s += "Converting these angles back to hkl resulted in hkl=" "(%f,%f,%f)" % (
                hkl[0],
                hkl[1],
                hkl[2],
            )
            raise DiffcalcException(s)

    def _verify_virtual_angles(
        self,
        h: float,
        k: float,
        l: float,
        pos: Position,
        virtual_angles: Dict[str, float],
    ) -> None:
        # Check that the virtual angles calculated/fixed during the hklToAngles
        # those read back from pos using anglesToVirtualAngles
        virtual_angles_readback = self.get_virtual_angles(pos, False)
        for key, val in virtual_angles.items():
            if val is not None:  # Some values calculated in some mode_selector
                r = virtual_angles_readback[key]
                if (not isnan(val) or not isnan(r)) and not radians_equivalent(
                    val, r, 1e-5
                ):
                    s = (
                        "ERROR: The angles calculated for hkl=(%f,%f,%f) with"
                        " mode=%s were %s.\n" % (h, k, l, self.__repr_mode(), str(pos))
                    )
                    s += (
                        "During verification the virtual angle %s resulting "
                        "from (or set for) this calculation of %f " % (key, val)
                    )
                    s += (
                        "did not match that calculated by "
                        "anglesToVirtualAngles of %f" % virtual_angles_readback[key]
                    )
                    raise DiffcalcException(s)



###
# Copyright 2008-2011 Diamond Light Source Ltd.
# This file is part of Diffcalc.
#
# Diffcalc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Diffcalc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Diffcalc.  If not, see <http://www.gnu.org/licenses/>.
###

from diffcalc.hkl.geometry import Position
from diffcalc.ub.reference import ReflectionList


class TestReflectionList:
    def setup_method(self):
        self.reflist = ReflectionList()
        pos = Position(0.1, 0.2, 0.3, 0.4, 0.5, 0.6)
        self.reflist.add_reflection((1, 2, 3), pos, 1000, "ref1")
        pos = Position(0.11, 0.22, 0.33, 0.44, 0.55, 0.66)
        self.reflist.add_reflection((1.1, 2.2, 3.3), pos, 1100, "ref2")

    def test_str(self):
        res = str(self.reflist)
        assert (
            res
            == """     ENERGY     H     K     L        MU    DELTA       NU      ETA      CHI      PHI  TAG
   1 1000.000  1.00  2.00  3.00    0.1000   0.2000   0.3000   0.4000   0.5000   0.6000  ref1
   2 1100.000  1.10  2.20  3.30    0.1100   0.2200   0.3300   0.4400   0.5500   0.6600  ref2"""
        )

    def test_add_reflection(self):
        assert len(self.reflist) == 2
        pos = Position(0.11, 0.22, 0.33, 0.44, 0.55, 0.66)
        self.reflist.add_reflection((11.1, 12.2, 13.3), pos, 1100, "ref2")

    def test_get_reflection(self):
        answered = self.reflist.get_reflection(1).astuple
        pos = Position(0.1, 0.2, 0.3, 0.4, 0.5, 0.6)
        desired = ((1, 2, 3), pos.astuple, 1000, "ref1")
        assert answered == desired
        answered = self.reflist.get_reflection("ref1").astuple
        assert answered == desired

    def test_remove_reflection(self):
        self.reflist.remove_reflection(1)
        answered = self.reflist.get_reflection(1).astuple
        pos = Position(0.11, 0.22, 0.33, 0.44, 0.55, 0.66)
        desired = ((1.1, 2.2, 3.3), pos.astuple, 1100, "ref2")
        assert answered == desired
        self.reflist.remove_reflection("ref2")
        assert self.reflist.reflections == []

    def test_edit_reflection(self):
        ps = Position(0.1, 0.2, 0.3, 0.4, 0.5, 0.6)
        self.reflist.edit_reflection(1, (10, 20, 30), ps, 1000, "new1")
        assert self.reflist.get_reflection(1).astuple == (
            (10, 20, 30),
            ps.astuple,
            1000,
            "new1",
        )
        pos = Position(0.11, 0.22, 0.33, 0.44, 0.55, 0.66)
        assert self.reflist.get_reflection(2).astuple == (
            (1.1, 2.2, 3.3),
            pos.astuple,
            1100,
            "ref2",
        )
        self.reflist.edit_reflection("ref2", (1.1, 2.2, 3.3), pos, 1100, "new2")
        assert self.reflist.get_reflection("new2").astuple == (
            (1.1, 2.2, 3.3),
            pos.astuple,
            1100,
            "new2",
        )
        self.reflist.edit_reflection("new2", (10, 20, 30), pos, 1100, "new1")
        assert self.reflist.get_reflection("new1").astuple == (
            (10, 20, 30),
            ps.astuple,
            1000,
            "new1",
        )

    def test_swap_reflection(self):
        self.reflist.swap_reflections(1, 2)
        pos = Position(0.11, 0.22, 0.33, 0.44, 0.55, 0.66)
        assert self.reflist.get_reflection(1).astuple == (
            (1.1, 2.2, 3.3),
            pos.astuple,
            1100,
            "ref2",
        )
        pos = Position(0.1, 0.2, 0.3, 0.4, 0.5, 0.6)
        assert self.reflist.get_reflection(2).astuple == (
            (1, 2, 3),
            pos.astuple,
            1000,
            "ref1",
        )
        self.reflist.swap_reflections("ref1", "ref2")
        pos = Position(0.11, 0.22, 0.33, 0.44, 0.55, 0.66)
        assert self.reflist.get_reflection(2).astuple == (
            (1.1, 2.2, 3.3),
            pos.astuple,
            1100,
            "ref2",
        )
        pos = Position(0.1, 0.2, 0.3, 0.4, 0.5, 0.6)
        assert self.reflist.get_reflection(1).astuple == (
            (1, 2, 3),
            pos.astuple,
            1000,
            "ref1",
        )



"""Routines for UB matrix calculation.

This module provides a number of objects and functions for setting UB matrix
and reference/surface normal vectors based on reflections and/or orientations
and crystal miscut information.
"""
import dataclasses
import pickle
import uuid
from copy import deepcopy
from itertools import product
from math import acos, asin, cos, degrees, pi, radians, sin
from typing import List, Optional, Sequence, Tuple, Union

import numpy as np
from diffcalc.hkl.geometry import Position, get_q_phi, get_rotation_matrices
from diffcalc.ub.crystal import Crystal
from diffcalc.ub.fitting import fit_crystal, fit_u_matrix
from diffcalc.ub.reference import OrientationList, Reflection, ReflectionList
from diffcalc.util import (
    SMALL,
    DiffcalcException,
    allnum,
    bound,
    cross3,
    dot3,
    is_small,
    xyz_rotation,
    zero_round,
)
from numpy.linalg import inv, norm


@dataclasses.dataclass
class ReferenceVector:
    """Class representing reference vector information.

    Reference vector object is used to define orientations
    in reciprocal and laboratory coordinate systems e.g. to
    represent an azimuthat direction or a surface normal.

    Attributes
    ----------
    n_ref: Tuple[float, float, float]
        tuple with vector coordinates
    rlv: bool
        Flag indicating if coordinates are in reciprocal or
        laboratory coordinate system.
        True: for reciprocal space vector.
        False: for real space vector.

    Methods
    -------
    get_array(UB: Optional[np.ndarray] = None) -> np.ndarray
        Return reference vector as 3x1 NumPy array
    set_array(n_ref: np.ndarray) -> Tuple[float, float, float]
        Set reference vector coordinates from 3x1 NumPy array
    """

    n_ref: Tuple[float, float, float]
    rlv: bool

    def get_array(self, UB: Optional[np.ndarray] = None) -> np.ndarray:
        """Return reference vector coordinates from (3, 1) NumPy array.

        Return reference vector coordinates in reciprocal of laboratory
        coordinate system. UB matrix is required when converting between
        alternative coordinate systems.

        Parameters
        ----------
        UB: np.ndarray, optional
            UB matrix as (3, 3) NumPy array object. Use UB matrix to convert
            coordinates between reciprocal and laboratory coordinate systems
            (default: None, do not convert coordinates into the alternative system)

        Returns
        -------
        np.ndarray
            Returns reference vector coordinates as (3, 1) NumPy array.

        Raises
        ------
        DiffcalcException
            If UB matrix is not a (3, 3) NumPy array.
        """
        n_ref_array = np.array([self.n_ref]).T
        if UB is None:
            return n_ref_array
        if not isinstance(UB, np.ndarray):
            raise DiffcalcException(
                "Invalid input parameter. UB must be a NumPy array object"
            )
        elif UB.shape != (3, 3):
            raise DiffcalcException(
                "Invalid array shape. UB array must have (3, 3) dimensions."
            )
        if self.rlv:
            n_ref_new = UB @ n_ref_array
        else:
            n_ref_new = inv(UB) @ n_ref_array
        return n_ref_new / norm(n_ref_new)

    def set_array(self, n_ref: np.ndarray) -> None:
        """Set reference vector coordinates from NumPy array.

        Parameters
        ----------
        n_ref: np.ndarray
            NumPy (3, 1) array with reference vector coordinates.

        Returns
        -------
        None

        Raises
        ------
        DiffcalcException
            If input parameter is not (3, 1) NumPy array

        """
        if not isinstance(n_ref, np.ndarray):
            raise DiffcalcException(
                "Input parameter in set_array method must be a NumPy array object."
            )
        if n_ref.shape != (3, 1):
            raise DiffcalcException(
                "Input NumPy array in set_array method must have (3, 1) shape."
            )
        (r1, r2, r3) = tuple(n_ref.T[0].tolist())
        self.n_ref = (r1, r2, r3)


class UBCalculation:
    """Class containing information required for for UB matrix calculation.

    Attributes
    ----------
    name: str, defalut = UUID
        Name for UB calculation. Default is to generate UUID value.
    crystal: Crystal
        Object containing crystal lattice parameters.
    reflist: ReflectionList
        List of reference reflections for UB matrix calculation.
    orientlist: OrientationList
        List of crystal orientations for UB matrix calculation.
    reference: ReferenceVector
        Object representing azimuthal reference vector.
    surface: ReferenceVector
        Object representing surface normal vector.
    U: np.ndarray
        U matrix as a NumPy array
    UB: np.ndarray
        UB matrix as a NumPy array
    """

    def __init__(self, name: Optional[str] = None) -> None:
        self.name: str = name if name is not None else str(uuid.uuid4())
        self.crystal: Crystal = None
        self.reflist: ReflectionList = ReflectionList()
        self.orientlist: OrientationList = OrientationList()
        self.reference: ReferenceVector = ReferenceVector((1, 0, 0), True)
        self.surface: ReferenceVector = ReferenceVector((0, 0, 1), False)
        self.U: np.ndarray = None
        self.UB: np.ndarray = None

        self._WIDTH: int = 13

    ### State ###
    @staticmethod
    def load(filename: str) -> "UBCalculation":
        """Load current UB matrix calculation from a pickle file.

        Parameters
        ----------
        filename : str
            Pickle file name.

        Returns
        -------
        UBCalculation
            UB calculation object.

        Raises
        ------
        DiffcalcException
            If object in pickle file isn't UBcalculation class instance.
        """
        with open(filename, "rb") as fp:
            obj = pickle.load(fp)
            if isinstance(obj, UBCalculation):
                return obj
            else:
                raise DiffcalcException(
                    f"Invalid object type {type(obj)} found in pickle file."
                )

    def pickle(self, filename: str) -> None:
        """Save current UB matrix calculation as a pickle file.

        Parameters
        ----------
        filename : str
            Pickle file name.
        """
        with open(filename, "wb") as fp:
            pickle.dump(obj=self, file=fp)

    def __str__(self) -> str:
        """Text representation of UB calculation object.

        Returns
        -------
        str
            String containing crystal lattice, reference vector, UB matrix
            and reference reflection/orientation information.
        """
        if self.name is None:
            return "<<< No UB calculation started >>>"
        lines = []
        lines.append("UBCALC")
        lines.append("")
        lines.append("   name:".ljust(self._WIDTH) + self.name.rjust(9))

        lines.append("")
        lines.append("REFERNCE")
        lines.append("")
        lines.extend(
            self.__str_lines_reference(self.n_hkl, self.n_phi, self.reference.rlv)
        )
        lines.append("")
        lines.append("SURFACE NORMAL")
        lines.append("")
        lines.extend(
            self.__str_lines_reference(self.surf_nhkl, self.surf_nphi, self.surface.rlv)
        )

        lines.append("")
        lines.append("CRYSTAL")
        lines.append("")

        if self.crystal is None:
            lines.append("   <<< none specified >>>")
        else:
            lines.append(str(self.crystal))

        lines.append("")
        lines.append("UB MATRIX")
        lines.append("")

        if self.UB is None:
            lines.append("   <<< none calculated >>>")
        else:
            lines.extend(self.__str_lines_u())
            lines.append("")
            lines.extend(self.__str_lines_ub_angle_and_axis())
            lines.append("")
            lines.extend(self.__str_lines_ub())

        lines.extend(self.__str_lines_refl())

        lines.extend(self.__str_lines_orient())

        return "\n".join(lines)

    def __str_lines_u(self) -> List[str]:
        lines: List[str] = []
        if self.U is None:
            return lines
        fmt = "% 9.5f % 9.5f % 9.5f"
        lines.append(
            "   U matrix:".ljust(self._WIDTH)
            + fmt
            % (
                zero_round(self.U[0, 0]),
                zero_round(self.U[0, 1]),
                zero_round(self.U[0, 2]),
            )
        )
        lines.append(
            " " * self._WIDTH
            + fmt
            % (
                zero_round(self.U[1, 0]),
                zero_round(self.U[1, 1]),
                zero_round(self.U[1, 2]),
            )
        )
        lines.append(
            " " * self._WIDTH
            + fmt
            % (
                zero_round(self.U[2, 0]),
                zero_round(self.U[2, 1]),
                zero_round(self.U[2, 2]),
            )
        )
        return lines

    def __str_lines_ub_angle_and_axis(self) -> List[str]:
        lines = []
        fmt = "% 9.5f % 9.5f % 9.5f"
        rotation_angle, rotation_axis = self.get_miscut()
        angle = 0 if is_small(rotation_angle) else degrees(rotation_angle)
        axis = tuple(0 if is_small(x) else x for x in rotation_axis.T.tolist()[0])
        if abs(norm(rotation_axis)) < SMALL:
            lines.append("   miscut angle:".ljust(self._WIDTH) + "  0")
        else:
            lines.append("   miscut:")
            lines.append("      angle:".ljust(self._WIDTH) + f"{angle:9.5f}")
            lines.append("       axis:".ljust(self._WIDTH) + fmt % axis)

        return lines

    def __str_lines_ub(self) -> List[str]:
        lines = []
        fmt = "% 9.5f % 9.5f % 9.5f"
        UB = self.UB
        lines.append(
            "   UB matrix:".ljust(self._WIDTH)
            + fmt % (zero_round(UB[0, 0]), zero_round(UB[0, 1]), zero_round(UB[0, 2]))
        )
        lines.append(
            " " * self._WIDTH
            + fmt % (zero_round(UB[1, 0]), zero_round(UB[1, 1]), zero_round(UB[1, 2]))
        )
        lines.append(
            " " * self._WIDTH
            + fmt % (zero_round(UB[2, 0]), zero_round(UB[2, 1]), zero_round(UB[2, 2]))
        )
        return lines

    def __str_vector(self, m: np.ndarray) -> str:
        if not isinstance(m, np.ndarray):
            return str(m)
        return " ".join([(f"{e:9.5f}").rjust(9) for e in m.T.tolist()[0]])

    def __str_lines_reference(
        self, ref_nhkl: np.ndarray, ref_nphi: np.ndarray, rlv: bool
    ) -> List[str]:
        SET_LABEL = " <- set"
        lines = []
        if rlv:
            nhkl_label = SET_LABEL
            nphi_label = ""
        else:
            nhkl_label = ""
            nphi_label = SET_LABEL

        lines.append(
            "   n_hkl:".ljust(self._WIDTH) + self.__str_vector(ref_nhkl) + nhkl_label
        )
        lines.append(
            "   n_phi:".ljust(self._WIDTH) + self.__str_vector(ref_nphi) + nphi_label
        )
        return lines

    def __str_lines_refl(self) -> List[str]:
        lines = ["", "REFLECTIONS", ""]

        lines.extend(self.reflist._str_lines())
        return lines

    def __str_lines_orient(self) -> List[str]:
        lines = ["", "CRYSTAL ORIENTATIONS", ""]

        lines.extend(self.orientlist._str_lines())
        return lines

    ### Lattice ###

    def set_lattice(
        self,
        name: str,
        system: Optional[
            Union[str, float]
        ] = None,  # FIXME: Cannot set Union type for  positional arguments
        a: Optional[float] = None,
        b: Optional[float] = None,
        c: Optional[float] = None,
        alpha: Optional[float] = None,
        beta: Optional[float] = None,
        gamma: Optional[float] = None,
    ) -> None:
        """Set crystal lattice parameters using shortform notation.

        Following combinations of system and lattice parameters are supported:

        ('Cubic', a) -- sets Cubic system
        ('Tetragonal', a, c) -- sets Tetragonal system
        ('Hexagonal', a, c) -- sets Hexagonal system
        ('Orthorhombic', a, b, c) -- sets Orthorombic system
        ('Rhombohedral', a, alpha) -- sets Rhombohedral system
        ('Monoclinic', a, b, c, beta) -- sets Monoclinic system
        ('Triclinic', a, b, c, alpha, beta, gamma) -- sets Triclinic system

        Crystal system can be inferred from the lattice parameters for the
        following cases:

        (a,) -- assumes Cubic system
        (a, c) -- assumes Tetragonal system
        (a, b, c) -- assumes Orthorombic system
        (a, b, c, angle) -- assumes Monoclinic system with beta not equal to 90 or
                            Hexagonal system if a = b and gamma = 120
        (a, b, c, alpha, beta, gamma) -- sets Triclinic system

        Parameters
        ----------
        name: str
            Crystal name
        system: Optional[float], default = None
            Crystal lattice type.
        a: Optional[float], default = None
            Crystal lattice parameter.
        b: Optional[float], default = None
            Crystal lattice parameter.
        c: Optional[float], default = None
            Crystal lattice parameter.
        alpha: Optional[float], default = None
            Crystal lattice angle.
        beta: Optional[float], default = None
            Crystal lattice angle.
        gamma: Optional[float], default = None
            Crystal lattice angle.
        """
        if not isinstance(name, str):
            raise TypeError("Invalid crystal name.")
        shortform = tuple(
            val for val in (system, a, b, c, alpha, beta, gamma) if val is not None
        )
        if not shortform:
            raise TypeError("Please specify unit cell parameters.")
        elif allnum(shortform):
            sf = shortform
            if len(sf) == 1:
                system = "Cubic"
            elif len(sf) == 2:
                system = "Tetragonal"
            elif len(sf) == 3:
                system = "Orthorhombic"
            elif len(sf) == 4:
                if is_small(float(sf[0]) - float(sf[1])) and sf[3] == 120:
                    system = "Hexagonal"
                else:
                    system = "Monoclinic"
            elif len(sf) == 6:
                system = "Triclinic"
            else:
                raise TypeError(
                    "Invalid number of input parameters to set unit lattice."
                )
            fullform = (system,) + shortform
        else:
            if not isinstance(shortform[0], str):
                raise TypeError("Invalid unit cell parameters specified.")
            fullform = shortform
        if self.name is None:
            raise DiffcalcException(
                "Cannot set lattice until a UBCalcaluation has been started "
                "with newubcalc"
            )
        self.crystal = Crystal(name, *fullform)

    ### Reference vector ###
    @property
    def n_hkl(self) -> np.ndarray:
        """Return reference vector property represented using miller indices.

        Returns
        -------
        np.ndarray:
            Reference vector represented as (3,1) NumPy array.
        """
        if self.UB is None and not self.reference.rlv:
            return None
        return self.reference.get_array(None if self.reference.rlv else self.UB)

    @n_hkl.setter
    def n_hkl(self, n_hkl: Tuple[float, float, float]) -> None:
        self.reference = ReferenceVector(deepcopy(n_hkl), True)

    @property
    def n_phi(self) -> np.ndarray:
        """Return reference vector property represented laboratory space coordinates.

        Returns
        -------
        np.ndarray:
            Reference vector represented as (3,1) NumPy array.
        """
        if self.UB is None and self.reference.rlv:
            return None
        return self.reference.get_array(self.UB if self.reference.rlv else None)

    @n_phi.setter
    def n_phi(self, n_phi: Tuple[float, float, float]) -> None:
        self.reference = ReferenceVector(deepcopy(n_phi), False)

    ### Surface vector ###
    @property
    def surf_nhkl(self) -> np.ndarray:
        """Return surface normal vector property represented using miller indices.

        Returns
        -------
        np.ndarray:
            Surface normal vector represented as (3,1) NumPy array.
        """
        return self.surface.get_array(None if self.surface.rlv else self.UB)

    @surf_nhkl.setter
    def surf_nhkl(self, surf_nhkl: Tuple[float, float, float]) -> None:
        self.surface = ReferenceVector(deepcopy(surf_nhkl), True)

    @property
    def surf_nphi(self) -> np.ndarray:
        """Return surface normal vector represented using laboratory space coordinates.

        Returns
        -------
        np.ndarray:
            Reference vector represented as (3,1) NumPy array.
        """
        return self.surface.get_array(self.UB if self.surface.rlv else None)

    @surf_nphi.setter
    def surf_nphi(self, surf_nphi: Tuple[float, float, float]) -> None:
        self.surface = ReferenceVector(deepcopy(surf_nphi), False)

    ### Reflections ###

    def add_reflection(
        self,
        hkl: Tuple[float, float, float],
        position: Position,
        energy: float,
        tag: Optional[str] = None,
    ):
        """Add a reference reflection.

        Adds a reflection position in degrees and in the systems internal
        representation.

        Parameters
        ----------
        hkl : Tuple[float, float, float]
            hkl index of the reflection
        position: Position
            list of diffractometer angles in internal representation in degrees
        energy : float
            energy of the x-ray beam
        tag : Optional[str], default = None
            identifying tag for the reflection
        """
        if self.reflist is None:
            raise DiffcalcException("No UBCalculation loaded")
        self.reflist.add_reflection(hkl, position, energy, tag)

    def edit_reflection(
        self,
        idx: Union[str, int],
        hkl: Tuple[float, float, float],
        position: Position,
        energy: float,
        tag: Optional[str] = None,
    ):
        """Change a reference reflection.

        Changes a reflection position in degrees and in the
        systems internal representation.

        Parameters
        ----------
        idx : Union[str, int]
            index or tag of the reflection to be changed
        hkl : Tuple[float, float, float]
            hkl index of the reflection
        position: Position
            list of diffractometer angles in internal representation in degrees
        energy : float
            energy of the x-ray beam
        tag : Optional[str], default = None
            identifying tag for the reflection
        """
        if self.reflist is None:
            raise DiffcalcException("No UBCalculation loaded")
        self.reflist.edit_reflection(idx, hkl, position, energy, tag)

    def get_reflection(self, idx: Union[str, int]) -> Reflection:
        """Get a reference reflection.

        Get a reflection position in degrees and in the
        systems internal representation.

        Parameters
        ----------
        idx : Union[str, int]
            index or tag of the reflection
        """
        return self.reflist.get_reflection(idx)

    def get_number_reflections(self) -> int:
        """Get a number of stored reference reflections.

        Returns
        -------
        int:
            Number of reference reflections.
        """
        try:
            return len(self.reflist)
        except TypeError:
            return 0

    def get_tag_refl_num(self, tag: str) -> int:
        """Get a reference reflection index.

        Get a reference reflection index for the
        provided reflection tag.

        Parameters
        ----------
        tag : str
            identifying tag for the reflection
        Returns
        -------
        int:
            Reference reflection index
        """
        if tag is None:
            raise IndexError("Reflection tag is None")
        return self.reflist.get_tag_index(tag) + 1

    def del_reflection(self, idx: Union[str, int]) -> None:
        """Delete a reference reflection.

        Parameters
        ----------
        idx : str or int
            index or tag of the deleted reflection
        """
        self.reflist.remove_reflection(idx)

    def swap_reflections(self, idx1: Union[str, int], idx2: Union[str, int]) -> None:
        """Swap indices of two reference reflections.

        Parameters
        ----------
        idx1 : str or int
            index or tag of the first reflection to be swapped
        idx2 : str or int
            index or tag of the second reflection to be swapped
        """
        self.reflist.swap_reflections(idx1, idx2)

    ### Orientations ###

    def add_orientation(self, hkl, xyz, position=None, tag=None):
        """Add a reference orientation.

        Adds a reference orientation in the diffractometer
        coordinate system.

        Parameters
        ----------
        hkl : :obj:`tuple` of numbers
            hkl index of the reference orientation
        xyz : :obj:`tuple` of numbers
            xyz coordinate of the reference orientation
        position: :obj:`list` or :obj:`tuple` of numbers
            list of diffractometer angles in internal representation in degrees
        tag : str
            identifying tag for the reflection
        """
        if self.orientlist is None:
            raise DiffcalcException("No UBCalculation loaded")
        if position is None:
            position = Position()
        self.orientlist.add_orientation(hkl, xyz, position, tag)

    def edit_orientation(self, idx, hkl, xyz, position=None, tag=None):
        """Change a reference orientation.

        Changes a reference orientation in the diffractometer
        coordinate system.

        Parameters
        ----------
        idx : str or int
            index or tag of the orientation to be changed
        hkl : :obj:`tuple` of numbers
            h index of the reference orientation
        xyz : :obj:`tuple` of numbers
            x coordinate of the reference orientation
        position: :obj:`list` or :obj:`tuple` of numbers
            list of diffractometer angles in internal representation in degrees
        tag : str
            identifying tag for the reflection
        """
        if self.orientlist is None:
            raise DiffcalcException("No UBCalculation loaded")
        self.orientlist.edit_orientation(
            idx, hkl, xyz, position if position is not None else Position(), tag
        )

    def get_orientation(self, idx):
        """Get a reference orientation.

        Get a reference orientation in the diffractometer
        coordinate system.

        Parameters
        ----------
        idx : str or int
            index or tag of the reference orientation
        """
        return self.orientlist.get_orientation(idx)

    def get_number_orientations(self):
        """Get a number of stored reference orientations.

        Returns
        -------
        int:
            Number of reference orientations.
        """
        try:
            return len(self.orientlist)
        except TypeError:
            return 0

    def get_tag_orient_num(self, tag):
        """Get a reference orientation index.

        Get a reference orientation index for the
        provided orientation tag.

        Parameters
        ----------
        tag : str
            identifying tag for the orientation
        Returns
        -------
        int:
            Reference orientation index
        """
        if tag is None:
            raise IndexError("Orientations tag is None")
        return self.orientlist.get_tag_index(tag) + 1

    def del_orientation(self, idx):
        """Delete a reference orientation.

        Parameters
        ----------
        idx : str or int
            index or tag of the deleted orientation
        """
        self.orientlist.remove_orientation(idx)

    def swap_orientations(self, idx1, idx2):
        """Swap indices of two reference orientations.

        Parameters
        ----------
        idx1 : str or int
            index or tag of the first orientation to be swapped
        idx2 : str or int
            index or tag of the second orientation to be swapped
        """
        self.orientlist.swap_orientations(idx1, idx2)

    ### Calculations ###

    def set_u(
        self,
        matrix: Union[
            np.ndarray,
            List[List[float]],
            Tuple[
                Tuple[float, float, float],
                Tuple[float, float, float],
                Tuple[float, float, float],
            ],
        ],
    ) -> None:
        """Manually sets U matrix.

        Set U matrix from input values. Update UB matrix if crystal lattice
        information is set.

        Parameters
        ----------
        matrix: Union(np.ndarray,
                      List[List[float]],
                      Tuple[Tuple[float, float, float],
                            Tuple[float, float, float],
                            Tuple[float, float, float]]
            Collection containing U matrix coordinates.

        Raises
        ------
        ValueError
            if collection isn't (3, 3) shape.
        """
        m = np.array(matrix, dtype=float)
        if len(m.shape) != 2 or m.shape[0] != 3 or m.shape[1] != 3:
            raise TypeError("set_u expects (3, 3) NumPy matrix.")

        if self.UB is None:
            print("Calculating UB matrix.")
        else:
            print("Recalculating UB matrix.")

        self.U = m
        if self.crystal is not None:
            self.UB = self.U @ self.crystal.B

    def set_ub(
        self,
        matrix: Union[
            np.ndarray,
            List[List[float]],
            Tuple[
                Tuple[float, float, float],
                Tuple[float, float, float],
                Tuple[float, float, float],
            ],
        ],
    ) -> None:
        """Manually sets UB matrix.

        Parameters
        ----------
        matrix: Union(np.ndarray,
                      List[List[float]],
                      Tuple[Tuple[float, float, float],
                            Tuple[float, float, float],
                            Tuple[float, float, float]]
            Collection containing UB matrix coordinates.

        Raises
        ------
        ValueError
            if collection isn't (3, 3) shape.
        """
        m = np.array(matrix, dtype=float)
        if len(m.shape) != 2 or m.shape[0] != 3 or m.shape[1] != 3:
            raise TypeError("set_ub expects (3, 3) NumPy matrix.")

        self.UB = m

    def _calc_ub_from_two_references(self, ref1, ref2):
        h1 = np.array([[ref1.h, ref1.k, ref1.l]]).T  # row->column
        h2 = np.array([[ref2.h, ref2.k, ref2.l]]).T

        # Compute the two reflections' reciprocal lattice vectors in the
        # cartesian crystal frame
        up = []
        for ref in (ref1, ref2):
            try:
                [MU, _, _, ETA, CHI, PHI] = get_rotation_matrices(ref.pos)
                Z = inv(PHI) @ inv(CHI) @ inv(ETA) @ inv(MU)
                up.append(Z @ np.array([[ref.x, ref.y, ref.z]]).T)
            except AttributeError:
                up.append(get_q_phi(ref.pos))
        u1p, u2p = up

        B = self.crystal.B
        h1c = B @ h1
        h2c = B @ h2

        # Create modified unit vectors t1, t2 and t3 in crystal and phi systems
        t1c = h1c
        t3c = cross3(h1c, h2c)
        t2c = cross3(t3c, t1c)

        t1p = u1p  # FIXED from h1c 9July08
        t3p = cross3(u1p, u2p)
        t2p = cross3(t3p, t1p)

        # ...and nornmalise and check that the reflections used are appropriate

        def __normalise(m):
            SMALL = 1e-4  # Taken from Vlieg's code
            d = norm(m)
            if d < SMALL:
                raise DiffcalcException(
                    "Invalid UB reference data. Please check that the specified "
                    "reference reflections/orientations are not parallel."
                )
            return m / d

        t1c = __normalise(t1c)
        t2c = __normalise(t2c)
        t3c = __normalise(t3c)

        t1p = __normalise(t1p)
        t2p = __normalise(t2p)
        t3p = __normalise(t3p)

        Tc = np.hstack([t1c, t2c, t3c])
        Tp = np.hstack([t1p, t2p, t3p])
        self.U = Tp @ inv(Tc)
        self.UB = self.U @ B

    def _get_calc_ub_references(self, idx1=None, idx2=None):
        try:
            ref1 = self.get_reflection(idx1)
        except Exception:
            try:
                ref1 = self.get_orientation(idx1)
            except Exception:
                raise DiffcalcException(
                    "Cannot find first reflection or orientation with index %s"
                    % str(idx1)
                )
        try:
            ref2 = self.get_reflection(idx2)
        except Exception:
            try:
                ref2 = self.get_orientation(idx2)
            except Exception:
                raise DiffcalcException(
                    "Cannot find second reflection or orientation with index %s"
                    % str(idx2)
                )
        return ref1, ref2

    def calc_ub(self, idx1=None, idx2=None):
        """Calculate UB matrix.

        Calculate UB matrix using two reference reflections and/or
        reference orientations.

        By default use the first two reference reflections when provided.
        If one or both reflections are not available use one or two reference
        orientations to complement mission reflection data.

        Parameters
        ----------
        idx1: int or str, optional
            The index or the tag of the first reflection or orientation.
        idx2: int or str, optional
            The index or the tag of the second reflection or orientation.
        """
        # Major variables:
        # h1, h2: user input reciprocal lattice vectors of the two reflections
        # h1c, h2c: user input vectors in cartesian crystal plane
        # pos1, pos2: measured diffractometer positions of the two reflections
        # u1a, u2a: measured reflection vectors in alpha frame
        # u1p, u2p: measured reflection vectors in phi frame

        # Get hkl and angle values for the first two reflections
        if self.crystal is None:
            raise DiffcalcException(
                "Not calculating UB matrix as no lattice parameters have "
                "been specified."
            )

        if idx1 is not None and idx2 is None:
            self._calc_ub_from_primary_only(idx1)
            return
        elif idx1 is None and idx2 is None:
            if self.get_number_reflections() == 1:
                self._calc_ub_from_primary_only()
                return
            ref_data = []
            for func, idx in product(
                (self.get_reflection, self.get_orientation), (1, 2)
            ):
                try:
                    ref_data.append(func(idx))
                except Exception:
                    pass
            try:
                ref1, ref2 = ref_data[:2]
            except ValueError:
                raise DiffcalcException(
                    "Cannot find reference vectors to calculate a U matrix. "
                    "Please add reference reflection and/or orientation data."
                )
        else:
            ref1, ref2 = self._get_calc_ub_references(idx1, idx2)

        self._calc_ub_from_two_references(ref1, ref2)

    def _calc_ub_from_primary_only(self, idx: int = 1) -> None:
        """Calculate orientation matrix from a single reference reflection.

        Parameters
        ----------
        idx: int
            Reference reflection index, default is 1.
        """
        # Algorithm from http://www.j3d.org/matrix_faq/matrfaq_latest.html

        # Get hkl and angle values for the first two reflections
        if self.crystal is None:
            raise DiffcalcException(
                "Not calculating UB matrix as no lattice parameters have "
                "been specified."
            )

        try:
            refl = self.get_reflection(idx)
        except IndexError:
            raise DiffcalcException(
                "One reflection is required to calculate a u matrix"
            )

        h = np.array([[refl.h], [refl.k], [refl.l]])
        B = self.crystal.B
        h_crystal = B @ h
        h_crystal = h_crystal * (1 / norm(h_crystal))

        q_measured_phi = get_q_phi(refl.pos)
        q_measured_phi = q_measured_phi * (1 / norm(q_measured_phi))

        rotation_axis = cross3(h_crystal, q_measured_phi)
        rotation_axis = rotation_axis * (1 / norm(rotation_axis))

        cos_rotation_angle = dot3(h_crystal, q_measured_phi)
        rotation_angle = acos(cos_rotation_angle)

        uvw = rotation_axis.T.tolist()[0]  # TODO: cleanup
        print(f"resulting U angle: {degrees(rotation_angle):.5f} deg")
        u_repr = ", ".join(["% .5f" % el for el in uvw])
        print("resulting U axis direction: [%s]" % u_repr)

        u, v, w = uvw
        rcos = cos(rotation_angle)
        rsin = sin(rotation_angle)
        m = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]  # TODO: tidy
        m[0][0] = rcos + u * u * (1 - rcos)
        m[1][0] = w * rsin + v * u * (1 - rcos)
        m[2][0] = -v * rsin + w * u * (1 - rcos)
        m[0][1] = -w * rsin + u * v * (1 - rcos)
        m[1][1] = rcos + v * v * (1 - rcos)
        m[2][1] = u * rsin + w * v * (1 - rcos)
        m[0][2] = v * rsin + u * w * (1 - rcos)
        m[1][2] = -u * rsin + v * w * (1 - rcos)
        m[2][2] = rcos + w * w * (1 - rcos)

        if self.UB is None:
            print("Calculating UB matrix from the first reflection only.")
        else:
            print("Recalculating UB matrix from the first reflection only.")
        print(
            "NOTE: A new UB matrix will not be automatically calculated "
            "when the orientation reflections are modified."
        )

        self.U = np.array(m)
        self.UB = self.U @ B

    def refine_ub(
        self,
        hkl: Tuple[float, float, float],
        position: Position,
        wavelength: float,
        refine_lattice=False,
        refine_umatrix=False,
    ):
        """Refine UB matrix to using single reflection.

        Refine UB matrix to match diffractometer position for the specified
        reflection. Refined U matrix will be accurate up to an azimuthal rotation
        around the specified scattering vector.

        Parameters
        ----------
        hkl: Tuple[float, float, float]
            Miller indices of the reflection.
        pos: Position
            Diffractometer position object.
        wavelength: float
            Radiation wavelength.
        refine_lattice: Optional[bool], default = False
            Apply refined lattice parameters to the current UB calculation object.
        refine_umatrix: Optional[bool], default = False
            Apply refined U matrix to the current UB calculation object.

        Returns
        -------
        Tuple[np.ndarray, Tuple[str, float, float, float, float, float, float]]
            Refined U matrix as NumPy array and refined crystal lattice parameters.
        """
        """
        refineub {[h k l]} {pos} -- refine unit cell dimensions and U matrix to match diffractometer angles for a given hkl value
        """
        scale, lat = self._rescale_unit_cell(hkl, position, wavelength)
        if scale and refine_lattice:
            self.set_lattice(*lat)
        mc_angle, mc_axis = self.get_miscut_from_hkl(hkl, position)
        if mc_angle and refine_umatrix:
            self.set_miscut(mc_axis, radians(mc_angle), True)

    def fit_ub(
        self,
        indices: Sequence[Union[str, int]],
        refine_lattice: Optional[bool] = False,
        refine_umatrix: Optional[bool] = False,
    ) -> Tuple[np.ndarray, Tuple[str, float, float, float, float, float, float]]:
        """Refine UB matrix using reference reflections.

        Parameters
        ----------
        indices: Sequence[Union[str, int]]
            List of reference reflection indices or tags.
        refine_lattice: Optional[bool], default = False
            Apply refined lattice parameters to the current UB calculation object.
        refine_umatrix: Optional[bool], default = False
            Apply refined U matrix to the current UB calculation object.

        Returns
        -------
        Tuple[np.ndarray, Tuple[str, float, float, float, float, float, float]]
            Refined U matrix as NumPy array and refined crystal lattice parameters.
        """
        if indices is None:
            raise DiffcalcException(
                "Please specify list of reference reflection indices."
            )
        if len(indices) < 3:
            raise DiffcalcException(
                "Need at least 3 reference reflections to fit UB matrix."
            )

        if len(self.crystal.get_lattice_params()[1]) == 6:
            new_u, new_lattice = self._fit_ub_uncon(indices)
        else:
            refl_list = []
            for idx in indices:
                try:
                    refl = self.get_reflection(idx)
                    refl_list.append(refl)
                except IndexError:
                    raise DiffcalcException(
                        "Cannot read reflection data for index %s" % str(idx)
                    )
            print("Fitting crystal lattice parameters...")
            new_crystal = fit_crystal(self.crystal, refl_list)
            print("Fitting orientation matrix...")
            new_u = fit_u_matrix(self.U, new_crystal, refl_list)
            new_lattice = (self.crystal.get_lattice()[0],) + new_crystal.get_lattice()[
                1:
            ]

        crystal_system = self.crystal.system

        if refine_lattice:
            self.set_lattice(new_lattice[0], crystal_system, *new_lattice[1:])

        if refine_umatrix:
            self.set_u(new_u)
        return new_u, new_lattice

    def _fit_ub_uncon(
        self, indices: Sequence[Union[str, int]]
    ) -> Tuple[np.ndarray, Tuple[str, float, float, float, float, float, float]]:
        """Refine UB matrix using least-squares solution."""
        if indices is None:
            raise DiffcalcException(
                "Please specify list of reference reflection indices."
            )
        if len(indices) < 3:
            raise DiffcalcException(
                "Need at least 3 reference reflections to fit UB matrix."
            )

        x = []
        y = []
        for idx in indices:
            try:
                refl = self.get_reflection(idx)
            except IndexError:
                raise DiffcalcException(
                    "Cannot read reflection data for index %s" % str(idx)
                )
            x.append((refl.h, refl.k, refl.l))
            wl = 12.3984 / refl.energy
            y_tmp = get_q_phi(refl.pos) * 2.0 * pi / wl
            y.append(y_tmp.T.tolist()[0])

        xm = np.array(x)
        ym = np.array(y)
        b = inv(xm.T @ xm) @ xm.T @ ym

        b1, b2, b3 = np.array([b[0]]), np.array([b[1]]), np.array([b[2]])
        e1 = b1 / norm(b1)
        e2 = b2 - e1 * dot3(b2.T, e1.T)
        e2 = e2 / norm(e2)
        e3 = b3 - e1 * dot3(b3.T, e1.T) - e2 * dot3(b3.T, e2.T)
        e3 = e3 / norm(e3)

        new_umatrix = np.array(e1.tolist() + e2.tolist() + e3.tolist()).T

        V = dot3(cross3(b1.T, b2.T), b3.T)
        a1 = cross3(b2.T, b3.T) * 2 * pi / V
        a2 = cross3(b3.T, b1.T) * 2 * pi / V
        a3 = cross3(b1.T, b2.T) * 2 * pi / V
        ax, bx, cx = norm(a1), norm(a2), norm(a3)
        alpha = acos(dot3(a2, a3) / (bx * cx))
        beta = acos(dot3(a1, a3) / (ax * cx))
        gamma = acos(dot3(a1, a2) / (ax * bx))

        lattice_name = self.crystal.get_lattice()[0]
        return new_umatrix, (
            lattice_name,
            ax,
            bx,
            cx,
            degrees(alpha),
            degrees(beta),
            degrees(gamma),
        )

    def get_miscut(self) -> Tuple[float, np.ndarray]:
        """Calculate miscut angle and axis from U matrix.

        Returns
        -------
        Tuple[float, np.ndarray]
            Miscut angle and miscut axis as (3,1) NumPy array.
        """
        surf_rot = self.U @ self.surf_nphi
        rotation_axis = cross3(self.surf_nphi, surf_rot)
        if abs(norm(rotation_axis)) < SMALL:
            rotation_axis = np.array([[0.0], [0.0], [0.0]])
            rotation_angle = 0.0
        else:
            rotation_axis = rotation_axis / norm(rotation_axis)
            cos_rotation_angle = bound(dot3(self.surf_nphi, surf_rot) / norm(surf_rot))
            rotation_angle = acos(cos_rotation_angle)
        return rotation_angle, rotation_axis

    def get_miscut_from_hkl(
        self, hkl: Tuple[float, float, float], pos: Position
    ) -> Tuple[float, Tuple[float, float, float]]:
        """Calculate miscut angle and axis from a single reflection.

        Using single reflection U matrix can be accurate up to an azimuthal
        rotation around the specified scattering vector.

        Parameters
        ----------
        hkl: Tuple[float, float, float]
            Miller indices of the reflection.
        pos: Position
            Diffractometer position object.

        Returns
        -------
        Tuple[float, Tuple[float, float, float]]
            The miscut angle and the corresponding miscut rotation axis.
        """
        q_vec = get_q_phi(pos)
        hkl_nphi = self.UB @ np.array([hkl]).T
        axis = cross3(hkl_nphi, q_vec)
        norm_axis = norm(axis)
        if norm_axis < SMALL:
            return None, None
        axis = axis / norm(axis)
        try:
            miscut = acos(bound(dot3(q_vec, hkl_nphi) / (norm(q_vec) * norm(hkl_nphi))))
        except AssertionError:
            return 0, (0, 0, 0)
        return degrees(miscut), (axis[0, 0], axis[1, 0], axis[2, 0])

    def set_miscut(
        self,
        xyz: Tuple[float, float, float],
        angle: float,
        add_miscut: Optional[bool] = False,
    ) -> None:
        """Calculate U matrix using a miscut axis and an angle.

        Parameters
        ----------
        xyz: Tuple[float, float, float]
            Rotation axis corresponding to the crystal miscut.
        angle: float
            The miscut angle.
        add_miscut: Optional[bool], default = False
            If False, set crystal miscut to the provided parameters.
            If True, apply provided miscut parameters to the existing settings.
        """
        if xyz is None:
            rot_matrix = xyz_rotation((0, 1, 0), angle)
        else:
            rot_matrix = xyz_rotation(xyz, angle)
        if self.U is not None and add_miscut:
            new_U = rot_matrix @ self.U
        else:
            new_U = rot_matrix
        self.set_u(new_U)

    def get_ttheta_from_hkl(self, hkl: Tuple[float, float, float], en: float) -> float:
        """Calculate two-theta scattering angle for a reflection.

        Parameters
        ----------
        hkl:  Tuple[float, float, float]
            Miller indices of the reflection.
        en: float
            Beam energy.

        Returns
        -------
        float
            Two-theta angle for the reflection.

        Raises
        ------
        ValueError
            If reflection is unreachable at the provided energy.
        """
        if self.crystal is None:
            raise DiffcalcException(
                "Cannot calculate theta angle as no lattice parameters have been specified."
            )
        wl = 12.39842 / en
        d = self.crystal.get_hkl_plane_distance(hkl)
        if wl > (2 * d):
            raise ValueError(
                "Reflection un-reachable as wavelength (%f) is more than twice\n"
                "the plane distance (%f)" % (wl, d)
            )
        try:
            return 2.0 * asin(wl / (d * 2))
        except ValueError as e:
            raise ValueError(
                f"asin(wl / (d * 2) with wl={wl:f} and d={d:f}: " + e.args[0]
            )

    def _rescale_unit_cell(
        self, hkl: Tuple[float, float, float], pos: Position, wavelength: float
    ) -> Tuple[float, Tuple[str, str, float, float, float, float, float, float]]:
        """Calculate unit cell scaling factor.

        Match lattice spacing corresponding to a given miller indices to the
        provided diffractometer position.

        Parameters
        ----------
        hkl: Tuple[float, float, float]
            Reflection mille indices.
        pos: Position
            Diffractometer position object.

        Returns
        -------
        Tuple[float, Tuple[str, str, float, float, float, float, float, float]]
            Scaling factor and updated crystal lattice parameters.
        """
        q_vec = get_q_phi(pos)
        q_hkl = norm(q_vec) / wavelength
        d_hkl = self.crystal.get_hkl_plane_distance(hkl)
        sc = 1 / (q_hkl * d_hkl)
        name, a1, a2, a3, alpha1, alpha2, alpha3 = self.crystal.get_lattice()
        if abs(sc - 1.0) < SMALL:
            return None, None
        h, k, l = hkl
        ref_a1 = sc * a1 if abs(h) > SMALL else a1
        ref_a2 = sc * a2 if abs(k) > SMALL else a2
        ref_a3 = sc * a3 if abs(l) > SMALL else a3
        return sc, (
            name,
            self.crystal.system,
            ref_a1,
            ref_a2,
            ref_a3,
            alpha1,
            alpha2,
            alpha3,
        )



###
# Copyright 2008-2011 Diamond Light Source Ltd.
# This file is part of Diffcalc.
#
# Diffcalc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Diffcalc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Diffcalc.  If not, see <http://www.gnu.org/licenses/>.
###

from math import pi

from diffcalc.ub.calc import UBCalculation
from numpy import array

from tests.diffcalc.scenarios import PosFromI16sEuler
from tests.test_tools import eq_
from tests.tools import matrixeq_

UB1 = (
    array(
        (
            (0.9996954135095477, -0.01745240643728364, -0.017449748351250637),
            (0.01744974835125045, 0.9998476951563913, -0.0003045864904520898),
            (0.017452406437283505, -1.1135499981271473e-16, 0.9998476951563912),
        )
    )
    * (2 * pi)
)

EN1 = 12.39842
REF1a = PosFromI16sEuler(1, 1, 30, 0, 60, 0)
REF1b = PosFromI16sEuler(1, 91, 30, 0, 60, 0)


def testAgainstI16Results():
    ubcalc = UBCalculation("cubcalc")
    ubcalc.set_lattice("latt", 1, 1, 1, 90, 90, 90)
    ubcalc.add_reflection((1, 0, 0), REF1a, EN1, "100")
    ubcalc.add_reflection((0, 0, 1), REF1b, EN1, "001")
    ubcalc.calc_ub()
    matrixeq_(ubcalc.UB, UB1)


def test_save_and_restore_empty_ubcalc(tmpdir):
    NAME = "test_save_and_restore_empty_ubcalc_with_one_already_started"
    ubcalc = UBCalculation(NAME)

    test_file = tmpdir / "test.pkl"
    ubcalc.pickle(test_file)

    ubcalc2 = UBCalculation.load(test_file)

    eq_(ubcalc2.name, NAME)


def test_save_and_restore_ubcalc_with_lattice(tmpdir):
    NAME = "test_save_and_restore_ubcalc_with_lattice"
    ubcalc = UBCalculation(NAME)
    ubcalc.set_lattice("latt", 1, 1, 1, 90, 90, 90)

    test_file = tmpdir / "test.pkl"
    ubcalc.pickle(test_file)

    ubcalc2 = UBCalculation.load(test_file)

    eq_(ubcalc2.crystal.get_lattice(), ("latt", 1, 1, 1, 90, 90, 90))


def test_save_and_restore_ubcalc_with_reflections(tmpdir):
    NAME = "test_save_and_restore_ubcalc_with_reflections"
    ubcalc = UBCalculation(NAME)
    ubcalc.add_reflection((1, 0, 0), REF1a, EN1, "100")
    ubcalc.add_reflection((0, 0, 1), REF1b, EN1, "001")
    ubcalc.add_reflection((0, 0, 1.5), REF1b, EN1, "001_5")
    ref1 = ubcalc.get_reflection(1)
    ref2 = ubcalc.get_reflection(2)
    ref3 = ubcalc.get_reflection(3)

    filename = tmpdir / "test_file.pkl"
    ubcalc.pickle(filename)

    ubcalc2 = UBCalculation.load(filename)

    eq_(ubcalc2.get_reflection(1).astuple, ref1.astuple)
    eq_(ubcalc2.get_reflection(2).astuple, ref2.astuple)
    eq_(ubcalc2.get_reflection(3).astuple, ref3.astuple)


def test_save_and_restore_ubcalc_with_UB_from_two_ref(tmpdir):
    NAME = "test_save_and_restore_ubcalc_with_UB_from_two_ref"
    ubcalc = UBCalculation(NAME)
    ubcalc.set_lattice("latt", 1, 1, 1, 90, 90, 90)
    ubcalc.add_reflection((1, 0, 0), REF1a, EN1, "100")
    ubcalc.add_reflection((0, 0, 1), REF1b, EN1, "001")
    ubcalc.calc_ub()
    matrixeq_(ubcalc.UB, UB1)

    filename = tmpdir / "test_file.pkl"
    ubcalc.pickle(filename)
    ubcalc2 = UBCalculation.load(filename)

    matrixeq_(ubcalc2.UB, ubcalc.UB)


def test_save_and_restore_ubcalc_with_UB_from_one_ref(tmpdir):
    NAME = "test_save_and_restore_ubcalc_with_UB_from_one_ref"
    ubcalc = UBCalculation(NAME)
    ubcalc.set_lattice("latt", 1, 1, 1, 90, 90, 90)
    ubcalc.add_reflection((1, 0, 0), REF1a, EN1, "100")
    ubcalc.calc_ub()
    matrixeq_(ubcalc.UB, UB1, places=2)

    filename = tmpdir / "test_file.pkl"
    ubcalc.pickle(filename)

    ubcalc2 = UBCalculation.load(filename)

    matrixeq_(ubcalc2.UB, ubcalc.UB)


def test_save_and_restore_ubcalc_with_manual_ub(tmpdir):
    NAME = "test_save_and_restore_ubcalc_with_manual_ub"
    UB = array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    ubcalc = UBCalculation(NAME)
    ubcalc.set_ub(UB)
    matrixeq_(ubcalc.UB, UB)

    filename = tmpdir / "test_file.pkl"
    ubcalc.pickle(filename)

    ubcalc2 = UBCalculation.load(filename)

    matrixeq_(ubcalc2.UB, ubcalc.UB)


def test_save_and_restore_ubcalc_with_manual_u(tmpdir):
    NAME = "test_save_and_restore_ubcalc_with_manual_u"
    U = array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    ubcalc = UBCalculation(NAME)
    ubcalc.set_lattice("latt", 1, 1, 1, 90, 90, 90)
    ubcalc.set_u(U)
    matrixeq_(ubcalc.UB, U * 2 * pi)
    filename = tmpdir / "test_file.pkl"
    ubcalc.pickle(filename)

    ubcalc2 = UBCalculation.load(filename)

    matrixeq_(ubcalc2.UB, ubcalc.UB)
